# -----------------------------------------------------------------------------

class DSType(object):
    """DS type."""
    HEP     = S_(DSHEP)
    NHEP    = S_(DSNHEP)
    GHEP    = S_(DSGHEP)
    GHIEP   = S_(DSGHIEP)
    GNHEP   = S_(DSGNHEP)
    NHEPTS  = S_(DSNHEPTS)
    SVD     = S_(DSSVD)
    HSVD    = S_(DSHSVD)
    GSVD    = S_(DSGSVD)
    PEP     = S_(DSPEP)
    NEP     = S_(DSNEP)

class DSStateType(object):
    """
    DS state types.

    - `RAW`:          Not processed yet.
    - `INTERMEDIATE`: Reduced to Hessenberg or tridiagonal form (or equivalent).
    - `CONDENSED`:    Reduced to Schur or diagonal form (or equivalent).
    - `TRUNCATED`:    Condensed form truncated to a smaller size.
    """
    RAW          = DS_STATE_RAW
    INTERMEDIATE = DS_STATE_INTERMEDIATE
    CONDENSED    = DS_STATE_CONDENSED
    TRUNCATED    = DS_STATE_TRUNCATED

class DSMatType(object):
    """
    To refer to one of the matrices stored internally in DS.

    - `A`:  first matrix of eigenproblem/singular value problem.
    - `B`:  second matrix of a generalized eigenproblem.
    - `C`:  third matrix of a quadratic eigenproblem.
    - `T`:  tridiagonal matrix.
    - `D`:  diagonal matrix.
    - `Q`:  orthogonal matrix of (right) Schur vectors.
    - `Z`:  orthogonal matrix of left Schur vectors.
    - `X`:  right eigenvectors.
    - `Y`:  left eigenvectors.
    - `U`:  left singular vectors.
    - `V`:  right singular vectors.
    - `W`:  workspace matrix.
    """
    A  = DS_MAT_A
    B  = DS_MAT_B
    C  = DS_MAT_C
    T  = DS_MAT_T
    D  = DS_MAT_D
    Q  = DS_MAT_Q
    Z  = DS_MAT_Z
    X  = DS_MAT_X
    Y  = DS_MAT_Y
    U  = DS_MAT_U
    V  = DS_MAT_V
    W  = DS_MAT_W

class DSParallelType(object):
    """
    DS parallel types.

    - `REDUNDANT`:    Every process performs the computation redundantly.
    - `SYNCHRONIZED`: The first process sends the result to the rest.
    - `DISTRIBUTED`:  Used in some cases to distribute the computation among
                      processes.
    """
    REDUNDANT    = DS_PARALLEL_REDUNDANT
    SYNCHRONIZED = DS_PARALLEL_SYNCHRONIZED
    DISTRIBUTED  = DS_PARALLEL_DISTRIBUTED

# -----------------------------------------------------------------------------

cdef class DS(Object):

    """DS."""

    Type         = DSType
    StateType    = DSStateType
    MatType      = DSMatType
    ParallelType = DSParallelType

    def __cinit__(self):
        self.obj = <PetscObject*> &self.ds
        self.ds = NULL

    def view(self, Viewer viewer=None) -> None:
        """
        Print the DS data structure.

        Collective.

        Parameters
        ----------
        viewer
            Visualization context; if not provided, the standard
            output is used.
        """
        cdef PetscViewer vwr = def_Viewer(viewer)
        CHKERR( DSView(self.ds, vwr) )

    def destroy(self) -> Self:
        """
        Destroy the DS object.

        Collective.
        """
        CHKERR( DSDestroy(&self.ds) )
        self.ds = NULL
        return self

    def reset(self) -> None:
        """
        Reset the DS object.

        Collective.
        """
        CHKERR( DSReset(self.ds) )

    def create(self, comm: Comm | None = None) -> Self:
        """
        Create the DS object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator; if not provided, it defaults to all processes.
        """
        cdef MPI_Comm ccomm = def_Comm(comm, SLEPC_COMM_DEFAULT())
        cdef SlepcDS newds = NULL
        CHKERR( DSCreate(ccomm, &newds) )
        CHKERR( SlepcCLEAR(self.obj) ); self.ds = newds
        return self

    def setType(self, ds_type: Type | str) -> None:
        """
        Set the type for the DS object.

        Logically collective.

        Parameters
        ----------
        ds_type
            The direct solver type to be used.
        """
        cdef SlepcDSType cval = NULL
        ds_type = str2bytes(ds_type, &cval)
        CHKERR( DSSetType(self.ds, cval) )

    def getType(self) -> str:
        """
        Get the DS type of this object.

        Not collective.

        Returns
        -------
        str
            The direct solver type currently being used.
        """
        cdef SlepcDSType ds_type = NULL
        CHKERR( DSGetType(self.ds, &ds_type) )
        return bytes2str(ds_type)

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Set the prefix used for searching for all DS options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all DS option requests.

        Notes
        -----
        A hyphen (``-``) must NOT be given at the beginning of the
        prefix name.  The first character of all runtime options is
        AUTOMATICALLY the hyphen.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( DSSetOptionsPrefix(self.ds, cval) )

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Append to the prefix used for searching for all DS options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all DS option requests.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( DSAppendOptionsPrefix(self.ds, cval) )

    def getOptionsPrefix(self) -> str:
        """
        Get the prefix used for searching for all DS options in the database.

        Not collective.

        Returns
        -------
        str
            The prefix string set for this DS object.
        """
        cdef const char *prefix = NULL
        CHKERR( DSGetOptionsPrefix(self.ds, &prefix) )
        return bytes2str(prefix)

    def setFromOptions(self) -> None:
        """
        Set DS options from the options database.

        Collective.

        Notes
        -----
        To see all options, run your program with the ``-help``
        option.
        """
        CHKERR( DSSetFromOptions(self.ds) )

    def duplicate(self) -> DS:
        """
        Duplicate the DS object with the same type and dimensions.

        Collective.
        """
        cdef DS ds = type(self)()
        CHKERR( DSDuplicate(self.ds, &ds.ds) )
        return ds

    #

    def allocate(self, ld: int) -> None:
        """
        Allocate memory for internal storage or matrices in DS.

        Logically collective.

        Parameters
        ----------
        ld
            Leading dimension (maximum allowed dimension for the
            matrices, including the extra row if present).
        """
        cdef PetscInt val = asInt(ld)
        CHKERR( DSAllocate(self.ds, val) )

    def getLeadingDimension(self) -> int:
        """
        Get the leading dimension of the allocated matrices.

        Not collective.

        Returns
        -------
        int
            Leading dimension (maximum allowed dimension for the matrices).
        """
        cdef PetscInt val = 0
        CHKERR( DSGetLeadingDimension(self.ds, &val) )
        return toInt(val)

    def setState(self, state: StateType) -> None:
        """
        Set the state of the DS object.

        Logically collective.

        Parameters
        ----------
        state
            The new state.

        Notes
        -----
        The state indicates that the dense system is in an initial
        state (raw), in an intermediate state (such as tridiagonal,
        Hessenberg or Hessenberg-triangular), in a condensed state
        (such as diagonal, Schur or generalized Schur), or in a
        truncated state.

        This function is normally used to return to the raw state when
        the condensed structure is destroyed.
        """
        cdef SlepcDSStateType val = state
        CHKERR( DSSetState(self.ds, val) )

    def getState(self) -> StateType:
        """
        Get the current state.

        Not collective.

        Returns
        -------
        StateType
            The current state.
        """
        cdef SlepcDSStateType val = DS_STATE_RAW
        CHKERR( DSGetState(self.ds, &val) )
        return val

    def setParallel(self, pmode: ParallelType) -> None:
        """
        Set the mode of operation in parallel runs.

        Logically collective.

        Parameters
        ----------
        pmode
            The parallel mode.
        """
        cdef SlepcDSParallelType val = pmode
        CHKERR( DSSetParallel(self.ds, val) )

    def getParallel(self) -> ParallelType:
        """
        Get the mode of operation in parallel runs.

        Not collective.

        Returns
        -------
        ParallelType
            The parallel mode.
        """
        cdef SlepcDSParallelType val = DS_PARALLEL_REDUNDANT
        CHKERR( DSGetParallel(self.ds, &val) )
        return val

    def setDimensions(self, n: int | None = None, l: int | None = None, k: int | None = None) -> None:
        """
        Set the matrices sizes in the DS object.

        Logically collective.

        Parameters
        ----------
        n
            The new size.
        l
            Number of locked (inactive) leading columns.
        k
            Intermediate dimension (e.g., position of arrow).

        Notes
        -----
        The internal arrays are not reallocated.
        """
        cdef PetscInt ival1 = PETSC_CURRENT
        cdef PetscInt ival2 = PETSC_CURRENT
        cdef PetscInt ival3 = PETSC_CURRENT
        if n is not None: ival1 = asInt(n)
        if l is not None: ival2 = asInt(l)
        if k is not None: ival3 = asInt(k)
        CHKERR( DSSetDimensions(self.ds, ival1, ival2, ival3) )

    def getDimensions(self) -> tuple[int, int, int, int]:
        """
        Get the current dimensions.

        Not collective.

        Returns
        -------
        n: int
            The new size.
        l: int
            Number of locked (inactive) leading columns.
        k: int
            Intermediate dimension (e.g., position of arrow).
        t: int
            Truncated length.
        """
        cdef PetscInt ival1 = 0
        cdef PetscInt ival2 = 0
        cdef PetscInt ival3 = 0
        cdef PetscInt ival4 = 0
        CHKERR( DSGetDimensions(self.ds, &ival1, &ival2, &ival3, &ival4) )
        return (toInt(ival1), toInt(ival2), toInt(ival3), toInt(ival4))

    def setBlockSize(self, bs: int) -> None:
        """
        Set the block size.

        Logically collective.

        Parameters
        ----------
        bs
            The block size.
        """
        cdef PetscInt val = bs
        CHKERR( DSSetBlockSize(self.ds, val) )

    def getBlockSize(self) -> int:
        """
        Get the block size.

        Not collective.

        Returns
        -------
        int
            The block size.
        """
        cdef PetscInt val = 0
        CHKERR( DSGetBlockSize(self.ds, &val) )
        return val

    def setMethod(self, meth: int) -> None:
        """
        Set the method to be used to solve the problem.

        Logically collective.

        Parameters
        ----------
        meth
            An index identifying the method.
        """
        cdef PetscInt val = meth
        CHKERR( DSSetMethod(self.ds, val) )

    def getMethod(self) -> int:
        """
        Get the method currently used in the DS.

        Not collective.

        Returns
        -------
        int
            Identifier of the method.
        """
        cdef PetscInt val = 0
        CHKERR( DSGetMethod(self.ds, &val) )
        return val

    def setCompact(self, comp: bool) -> None:
        """
        Set the matrices' compact storage flag.

        Logically collective.

        Parameters
        ----------
        comp
            True means compact storage.

        Notes
        -----
        Compact storage is used in some `DS` types such as
        `DS.Type.HEP` when the matrix is tridiagonal. This flag
        can be used to indicate whether the user provides the
        matrix entries via the compact form (the tridiagonal
        `DS.MatType.T`) or the non-compact one (`DS.MatType.A`).

        The default is ``False``.
        """
        cdef PetscBool val = asBool(comp)
        CHKERR( DSSetCompact(self.ds, val) )

    def getCompact(self) -> bool:
        """
        Get the compact storage flag.

        Not collective.

        Returns
        -------
        bool
            The flag.
        """
        cdef PetscBool val = PETSC_FALSE
        CHKERR( DSGetCompact(self.ds, &val) )
        return toBool(val)

    def setExtraRow(self, ext: bool) -> None:
        """
        Set a flag to indicate that the matrix has one extra row.

        Logically collective.

        Parameters
        ----------
        ext
            True if the matrix has extra row.

        Notes
        -----
        In Krylov methods it is useful that the matrix representing the direct
        solver has one extra row, i.e., has dimension :math:`(n+1) n`. If
        this flag is activated, all transformations applied to the right of the
        matrix also affect this additional row. In that case, :math:`(n+1)`
        must be less or equal than the leading dimension.

        The default is ``False``.
        """
        cdef PetscBool val = asBool(ext)
        CHKERR( DSSetExtraRow(self.ds, val) )

    def getExtraRow(self) -> bool:
        """
        Get the extra row flag.

        Not collective.

        Returns
        -------
        bool
            The flag.
        """
        cdef PetscBool val = PETSC_FALSE
        CHKERR( DSGetExtraRow(self.ds, &val) )
        return toBool(val)

    def setRefined(self, ref: bool) -> None:
        """
        Set a flag to indicate that refined vectors must be computed.

        Logically collective.

        Parameters
        ----------
        ref
            True if refined vectors must be used.

        Notes
        -----
        Normally the vectors returned in `DS.MatType.X` are eigenvectors of
        the projected matrix. With this flag activated, `vectors()` will return
        the right singular vector of the smallest singular value of matrix
        :math:`At - theta I`, where :math:`At` is the extended
        :math:`(n+1) times n` matrix and :math:`theta` is the Ritz value.
        This is used in the refined Ritz approximation.

        The default is ``False``.
        """
        cdef PetscBool val = asBool(ref)
        CHKERR( DSSetRefined(self.ds, val) )

    def getRefined(self) -> bool:
        """
        Get the refined vectors flag.

        Not collective.

        Returns
        -------
        bool
            The flag.
        """
        cdef PetscBool val = PETSC_FALSE
        CHKERR( DSGetRefined(self.ds, &val) )
        return toBool(val)

    def truncate(self, n: int, trim: bool = False) -> None:
        """
        Truncate the system represented in the DS object.

        Logically collective.

        Parameters
        ----------
        n
            The new size.
        trim
            A flag to indicate if the factorization must be trimmed.
        """
        cdef PetscInt val = asInt(n)
        cdef PetscBool flg = asBool(trim)
        CHKERR( DSTruncate(self.ds, val, flg) )

    def updateExtraRow(self) -> None:
        """
        Ensure that the extra row gets up-to-date after a call to `DS.solve()`.

        Logically collective.

        Perform all necessary operations so that the extra row gets up-to-date
        after a call to `DS.solve()`.
        """
        CHKERR( DSUpdateExtraRow(self.ds) )

    def getMat(self, matname: MatType) -> petsc4py.PETSc.Mat:
        """
        Get the requested matrix as a sequential dense Mat object.

        Not collective.

        Parameters
        ----------
        matname
            The requested matrix.

        Returns
        -------
        petsc4py.PETSc.Mat
            The matrix.
        """
        cdef SlepcDSMatType mname = matname
        cdef Mat mat = Mat()
        CHKERR( DSGetMat(self.ds, mname, &mat.mat) )
        CHKERR( PetscINCREF(mat.obj) )
        return mat

    def restoreMat(self, matname: MatType, Mat mat: petsc4py.PETSc.Mat) -> None:
        """
        Restore the previously seized matrix.

        Not collective.

        Parameters
        ----------
        matname
            The selected matrix.
        mat
            The matrix previously obtained with `getMat()`.
        """
        cdef SlepcDSMatType mname = matname
        CHKERR( PetscObjectDereference(<PetscObject>mat.mat) )
        CHKERR( DSRestoreMat(self.ds, mname, &mat.mat) )

    def setIdentity(self, matname: MatType) -> None:
        """
        Set the identity on the active part of a matrix.

        Logically collective.

        Parameters
        ----------
        matname
            The requested matrix.
        """
        cdef SlepcDSMatType mname = matname
        CHKERR( DSSetIdentity(self.ds, mname) )

    #

    def cond(self) -> float:
        """
        Compute the inf-norm condition number of the first matrix.

        Logically collective.

        Returns
        -------
        float
            Condition number.
        """
        cdef PetscReal rval = 0
        CHKERR( DSCond(self.ds, &rval) )
        return toReal(rval)

    def solve(self) -> ArrayScalar:
        """
        Solve the problem.

        Logically collective.

        Returns
        -------
        ArrayScalar
            Eigenvalues or singular values.
        """
        n = self.getDimensions()[0]
        cdef PetscScalar *eigr = NULL
        cdef PetscScalar *eigi = NULL
        cdef tmpr = allocate(<size_t>n*sizeof(PetscScalar), <void**>&eigr)
        cdef tmpi = allocate(<size_t>n*sizeof(PetscScalar), <void**>&eigi)
        CHKERR( DSSolve(self.ds, eigr, eigi) )
        cdef object kr = array_s(n, eigr)
        cdef object ki = array_s(n, eigi)
        if self.getType().upper() in ['HEP','GHEP','BSE','SVD','HSVD','GSVD']:
            return kr.real.copy()
        else:
            return kr+1j*ki

    def vectors(self, matname = MatType.X) -> None:
        """
        Compute vectors associated to the dense system such as eigenvectors.

        Logically collective.

        Parameters
        ----------
        matname: `DS.MatType` enumerate
           The matrix, used to indicate which vectors are required.
        """
        cdef SlepcDSMatType mname = matname
        CHKERR( DSVectors(self.ds, mname, NULL, NULL) )

    #

    def setSVDDimensions(self, m: int) -> None:
        """
        Set the number of columns of a `DS` of type `SVD`.

        Logically collective.

        Parameters
        ----------
        m
            The number of columns.
        """
        cdef PetscInt val = asInt(m)
        CHKERR( DSSVDSetDimensions(self.ds, val) )

    def getSVDDimensions(self) -> int:
        """
        Get the number of columns of a `DS` of type `SVD`.

        Not collective.

        Returns
        -------
        int
            The number of columns.
        """
        cdef PetscInt val = 0
        CHKERR( DSSVDGetDimensions(self.ds, &val) )
        return toInt(val)

    def setHSVDDimensions(self, m: int) -> None:
        """
        Set the number of columns of a `DS` of type `HSVD`.

        Logically collective.

        Parameters
        ----------
        m
            The number of columns.
        """
        cdef PetscInt val = asInt(m)
        CHKERR( DSHSVDSetDimensions(self.ds, val) )

    def getHSVDDimensions(self) -> int:
        """
        Get the number of columns of a `DS` of type `HSVD`.

        Not collective.

        Returns
        -------
        int
            The number of columns.
        """
        cdef PetscInt val = 0
        CHKERR( DSHSVDGetDimensions(self.ds, &val) )
        return toInt(val)

    def setGSVDDimensions(self, m: int, p: int) -> None:
        """
        Set the number of columns and rows of a `DS` of type `GSVD`.

        Logically collective.

        Parameters
        ----------
        m
            The number of columns.
        p
            The number of rows for the second matrix.
        """
        cdef PetscInt val1 = asInt(m)
        cdef PetscInt val2 = asInt(p)
        CHKERR( DSGSVDSetDimensions(self.ds, val1, val2) )

    def getGSVDDimensions(self) -> tuple[int, int]:
        """
        Get the number of columns and rows of a `DS` of type `GSVD`.

        Not collective.

        Returns
        -------
        m: int
            The number of columns.
        p: int
            The number of rows for the second matrix.
        """
        cdef PetscInt val1 = 0
        cdef PetscInt val2 = 0
        CHKERR( DSGSVDGetDimensions(self.ds, &val1, &val2) )
        return (toInt(val1), toInt(val2))

    def setPEPDegree(self, deg: int) -> None:
        """
        Set the polynomial degree of a `DS` of type `PEP`.

        Logically collective.

        Parameters
        ----------
        deg
            The polynomial degree.
        """
        cdef PetscInt val = asInt(deg)
        CHKERR( DSPEPSetDegree(self.ds, val) )

    def getPEPDegree(self) -> int:
        """
        Get the polynomial degree of a `DS` of type `PEP`.

        Not collective.

        Returns
        -------
        int
            The polynomial degree.
        """
        cdef PetscInt val = 0
        CHKERR( DSPEPGetDegree(self.ds, &val) )
        return toInt(val)

    def setPEPCoefficients(self, pbc: Sequence[float]) -> None:
        """
        Set the polynomial basis coefficients of a `DS` of type `PEP`.

        Logically collective.

        Parameters
        ----------
        pbc
            Coefficients.
        """
        cdef PetscInt na = 0
        cdef PetscReal *a = NULL
        cdef object tmp1 = iarray_r(pbc, &na, &a)
        CHKERR( DSPEPSetCoefficients(self.ds, a) )

    def getPEPCoefficients(self) -> ArrayReal:
        """
        Get the polynomial basis coefficients of a `DS` of type `PEP`.

        Not collective.

        Returns
        -------
        ArrayReal
            Coefficients.
        """
        cdef PetscInt np = 0
        cdef PetscReal *coeff = NULL
        CHKERR( DSPEPGetDegree(self.ds, &np) )
        CHKERR( DSPEPGetCoefficients(self.ds, &coeff) )
        cdef object ocoeff = None
        try:
            ocoeff = array_r(3*(np+1), coeff)
        finally:
            CHKERR( PetscFree(coeff) )
        return ocoeff

    #

    property state:
        """The state of the DS object."""
        def __get__(self) -> DSStateType:
            return self.getState()
        def __set__(self, value):
            self.setState(value)

    property parallel:
        """The mode of operation in parallel runs."""
        def __get__(self) -> DSParallelType:
            return self.getParallel()
        def __set__(self, value):
            self.setParallel(value)

    property block_size:
        """The block size."""
        def __get__(self) -> int:
            return self.getBlockSize()
        def __set__(self, value):
            self.setBlockSize(value)

    property method:
        """The method to be used to solve the problem."""
        def __get__(self) -> int:
            return self.getMethod()
        def __set__(self, value):
            self.setMethod(value)

    property compact:
        """Compact storage of matrices."""
        def __get__(self) -> bool:
            return self.getCompact()
        def __set__(self, value):
            self.setCompact(value)

    property extra_row:
        """If the matrix has one extra row."""
        def __get__(self) -> bool:
            return self.getExtraRow()
        def __set__(self, value):
            self.setExtraRow(value)

    property refined:
        """If refined vectors must be computed."""
        def __get__(self) -> bool:
            return self.getRefined()
        def __set__(self, value):
            self.setRefined(value)

# -----------------------------------------------------------------------------

del DSType
del DSStateType
del DSMatType
del DSParallelType

# -----------------------------------------------------------------------------



"""Type stubs for SLEPc Sys module."""

class Sys:
    """SLEPc system utilities."""

    @classmethod
    def getVersion(
        cls,
        devel: bool = False,
        date: bool = False,
        author: bool = False,
    ) -> tuple[int, int, int] | tuple[tuple[int, int, int], ...]:
        """
        Return SLEPc version information.

        Parameters
        ----------
        devel
            Additionally, return whether using an in-development version.
        date
            Additionally, return date information.
        author
            Additionally, return author information.

        Returns
        -------
        major : int
            Major version number.
        minor : int
            Minor version number.
        micro : int
            Micro (or patch) version number.
        """
        ...

    @classmethod
    def getVersionInfo(cls) -> dict[str, bool | int | str | tuple[str, ...]]:
        """
        Return SLEPc version information.

        Returns
        -------
        dict
            Dictionary with version information including:
            major, minor, subminor, release, date, authorinfo.
        """
        ...

    @classmethod
    def isInitialized(cls) -> bool:
        """
        Return whether SLEPc has been initialized.

        Returns
        -------
        bool
            True if SLEPc has been initialized.
        """
        ...

    @classmethod
    def isFinalized(cls) -> bool:
        """
        Return whether SLEPc has been finalized.

        Returns
        -------
        bool
            True if SLEPc has been finalized.
        """
        ...

    @classmethod
    def hasExternalPackage(cls, package: str) -> bool:
        """
        Return whether SLEPc has support for external package.

        Parameters
        ----------
        package
            The external package name.

        Returns
        -------
        bool
            True if SLEPc has support for the external package.
        """
        ...



# --------------------------------------------------------------------

cdef extern from * nogil:
    """
    #include "pyapicompat.h"
    #include "lib-slepc/compat.h"
    #include "lib-slepc/custom.h"

    /* Silence Clang warnings in Cython-generated C code */
    #if defined(__clang__)
      #pragma clang diagnostic ignored "-Wextra-semi-stmt"
      #pragma clang diagnostic ignored "-Wparentheses-equality"
      #pragma clang diagnostic ignored "-Wunreachable-code-fallthrough"
      #pragma clang diagnostic ignored "-Woverlength-strings"
      #pragma clang diagnostic ignored "-Wunreachable-code"
      #pragma clang diagnostic ignored "-Wundef"
    #elif defined(__GNUC__) || defined(__GNUG__)
      #pragma GCC diagnostic ignored "-Wstrict-aliasing"
      #pragma GCC diagnostic ignored "-Wtype-limits"
    #endif
    """

# -----------------------------------------------------------------------------

from petsc4py.PETSc import COMM_NULL
from petsc4py.PETSc import COMM_SELF
from petsc4py.PETSc import COMM_WORLD

# -----------------------------------------------------------------------------

from petsc4py.PETSc cimport MPI_Comm
from petsc4py.PETSc cimport PetscErrorCode, PetscErrorType
from petsc4py.PETSc cimport PETSC_SUCCESS, PETSC_ERR_PYTHON
from petsc4py.PETSc cimport CHKERR
from petsc4py.PETSc cimport PetscObject, PetscViewer
from petsc4py.PETSc cimport PetscRandom
from petsc4py.PETSc cimport PetscVec, PetscMat
from petsc4py.PETSc cimport PetscKSP, PetscPC

from petsc4py.PETSc cimport Comm
from petsc4py.PETSc cimport Object, Viewer
from petsc4py.PETSc cimport Random
from petsc4py.PETSc cimport Vec, Mat
from petsc4py.PETSc cimport KSP, PC

# -----------------------------------------------------------------------------

cdef inline object bytes2str(const char p[]):
     if p == NULL:
         return None
     cdef bytes s = <char*>p
     if isinstance(s, str):
         return s
     else:
         return s.decode()

cdef inline object str2bytes(object s, const char *p[]):
    if s is None:
        p[0] = NULL
        return None
    if not isinstance(s, bytes):
        s = s.encode()
    p[0] = <const char*>(<char*>s)
    return s

cdef inline object S_(const char p[]):
     if p == NULL: return None
     cdef object s = <char*>p
     return s if isinstance(s, str) else s.decode()

include "allocate.pxi"

# -----------------------------------------------------------------------------

cdef extern from * nogil:
    ctypedef bint      PetscBool
    const    PetscBool PETSC_TRUE
    const    PetscBool PETSC_FALSE
    ctypedef long      PetscInt
    ctypedef double    PetscReal
    ctypedef double    PetscScalar

cdef inline object toBool(PetscBool value):
    return True if value == PETSC_TRUE else False
cdef inline PetscBool asBool(object value) except? <PetscBool>0:
    return PETSC_TRUE if value else PETSC_FALSE

cdef inline object toInt(PetscInt value):
    return value
cdef inline PetscInt asInt(object value) except? -1:
    return value

cdef inline object toReal(PetscReal value):
    return value
cdef inline PetscReal asReal(object value) except? -1:
    return value

cdef extern from "<petsc4py/pyscalar.h>":
    object      PyPetscScalar_FromPetscScalar(PetscScalar)
    PetscScalar PyPetscScalar_AsPetscScalar(object) except? <PetscScalar>-1.0

cdef inline object toScalar(PetscScalar value):
    return PyPetscScalar_FromPetscScalar(value)
cdef inline PetscScalar asScalar(object value) except? <PetscScalar>-1.0:
    return PyPetscScalar_AsPetscScalar(value)

cdef extern from "Python.h":
     double PyComplex_RealAsDouble(object)
     double PyComplex_ImagAsDouble(object)

cdef inline object toComplex(PetscScalar rvalue, PetscScalar ivalue):
    return complex(toScalar(rvalue), toScalar(ivalue))

cdef inline PetscReal asComplexReal(object value) except? <PetscReal>-1.0:
    return <PetscReal>PyComplex_RealAsDouble(value)

cdef inline PetscReal asComplexImag(object value) except? <PetscReal>-1.0:
    return <PetscReal>PyComplex_ImagAsDouble(value)

cdef extern from * nogil:
    PetscReal PetscRealPart(PetscScalar v)
    PetscReal PetscImaginaryPart(PetscScalar v)

# --------------------------------------------------------------------

# NumPy support
# -------------

include "arraynpy.pxi"

import_array()

IntType     = PyArray_TypeObjectFromType(NPY_PETSC_INT)
RealType    = PyArray_TypeObjectFromType(NPY_PETSC_REAL)
ScalarType  = PyArray_TypeObjectFromType(NPY_PETSC_SCALAR)
ComplexType = PyArray_TypeObjectFromType(NPY_PETSC_COMPLEX)

# -----------------------------------------------------------------------------

cdef extern from "<string.h>"  nogil:
    void* memset(void*,int,size_t)
    void* memcpy(void*,void*,size_t)
    char* strdup(char*)

# -----------------------------------------------------------------------------

include "slepcmpi.pxi"
include "slepcsys.pxi"
include "slepcutil.pxi"
include "slepcst.pxi"
include "slepcbv.pxi"
include "slepcds.pxi"
include "slepcfn.pxi"
include "slepcrg.pxi"
include "slepceps.pxi"
include "slepcsvd.pxi"
include "slepcpep.pxi"
include "slepcnep.pxi"
include "slepcmfn.pxi"
include "slepclme.pxi"

# -----------------------------------------------------------------------------

__doc__ = u"""
Scalable Library for Eigenvalue Problem Computations
"""

DECIDE    = PETSC_DECIDE
DEFAULT   = PETSC_DEFAULT
DETERMINE = PETSC_DETERMINE
CURRENT   = PETSC_CURRENT

include "Sys.pyx"
include "Util.pyx"
include "ST.pyx"
include "BV.pyx"
include "DS.pyx"
include "FN.pyx"
include "RG.pyx"
include "EPS.pyx"
include "SVD.pyx"
include "PEP.pyx"
include "NEP.pyx"
include "MFN.pyx"
include "LME.pyx"

# -----------------------------------------------------------------------------

include "CAPI.pyx"

# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    int Py_AtExit(void (*)() noexcept nogil)
    void PySys_WriteStderr(char*,...)

cdef extern from "<stdio.h>" nogil:
    ctypedef struct FILE
    FILE *stderr
    int fprintf(FILE *, char *, ...)

cdef int initialize(object args) except PETSC_ERR_PYTHON:
    if (<int>SlepcInitializeCalled): return 1
    if (<int>SlepcFinalizeCalled):   return 0
    # initialize SLEPC
    CHKERR( SlepcInitialize(NULL, NULL, NULL, NULL) )
    # register finalization function
    if Py_AtExit(finalize) < 0:
        PySys_WriteStderr(b"warning: could not register %s with Py_AtExit()",
                          b"SlepcFinalize()")
    return 1 # and we are done, enjoy !!

from petsc4py.PETSc cimport PyPetscType_Register

cdef extern from * nogil:
    PetscErrorCode SlepcInitializePackageAll()
    ctypedef int PetscClassId
    PetscClassId SLEPC_ST_CLASSID  "ST_CLASSID"
    PetscClassId SLEPC_BV_CLASSID  "BV_CLASSID"
    PetscClassId SLEPC_DS_CLASSID  "DS_CLASSID"
    PetscClassId SLEPC_FN_CLASSID  "FN_CLASSID"
    PetscClassId SLEPC_RG_CLASSID  "RG_CLASSID"
    PetscClassId SLEPC_EPS_CLASSID "EPS_CLASSID"
    PetscClassId SLEPC_SVD_CLASSID "SVD_CLASSID"
    PetscClassId SLEPC_PEP_CLASSID "PEP_CLASSID"
    PetscClassId SLEPC_NEP_CLASSID "NEP_CLASSID"
    PetscClassId SLEPC_MFN_CLASSID "MFN_CLASSID"
    PetscClassId SLEPC_LME_CLASSID "LME_CLASSID"

cdef PetscErrorCode register() except PETSC_ERR_PYTHON:
    # make sure all SLEPc packages are initialized
    CHKERR( SlepcInitializePackageAll() )
    # register Python types
    PyPetscType_Register(SLEPC_ST_CLASSID,  ST)
    PyPetscType_Register(SLEPC_BV_CLASSID,  BV)
    PyPetscType_Register(SLEPC_DS_CLASSID,  DS)
    PyPetscType_Register(SLEPC_FN_CLASSID,  FN)
    PyPetscType_Register(SLEPC_RG_CLASSID,  RG)
    PyPetscType_Register(SLEPC_EPS_CLASSID, EPS)
    PyPetscType_Register(SLEPC_SVD_CLASSID, SVD)
    PyPetscType_Register(SLEPC_PEP_CLASSID, PEP)
    PyPetscType_Register(SLEPC_NEP_CLASSID, NEP)
    PyPetscType_Register(SLEPC_MFN_CLASSID, MFN)
    PyPetscType_Register(SLEPC_LME_CLASSID, LME)
    return PETSC_SUCCESS

cdef void finalize() noexcept nogil:
    cdef PetscErrorCode ierr = PETSC_SUCCESS
    # manage SLEPc finalization
    if not (<int>SlepcInitializeCalled): return
    if (<int>SlepcFinalizeCalled): return
    # finalize SLEPc
    ierr = SlepcFinalize()
    if ierr != PETSC_SUCCESS:
        fprintf(stderr, "SlepcFinalize() failed "
                "[error code: %d]\n", <int>ierr)
    # and we are done, see you later !!

# -----------------------------------------------------------------------------

def _initialize(args=None):
    cdef int ready = initialize(args)
    if ready: register()

def _finalize():
    finalize()

# -----------------------------------------------------------------------------



"""Type stubs for SLEPc FN module."""

from enum import IntEnum, StrEnum
from typing import Self, Sequence

from petsc4py.PETSc import Comm, Mat, Vec, Viewer
from petsc4py.typing import ArrayScalar, Scalar

class FNType(StrEnum):
    """FN type."""

    COMBINE = ...
    """Combination of functions."""
    RATIONAL = ...
    """Rational function."""
    EXP = ...
    """Exponential function."""
    LOG = ...
    """Logarithm function."""
    PHI = ...
    """Phi-function."""
    SQRT = ...
    """Square root function."""
    INVSQRT = ...
    """Inverse square root function."""

class FNCombineType(IntEnum):
    """FN type of combination of child functions.

    - `ADD`:       Addition         f(x) = f1(x)+f2(x)
    - `MULTIPLY`:  Multiplication   f(x) = f1(x)*f2(x)
    - `DIVIDE`:    Division         f(x) = f1(x)/f2(x)
    - `COMPOSE`:   Composition      f(x) = f2(f1(x))
    """

    ADD = ...
    """Addition f(x) = f1(x)+f2(x)."""
    MULTIPLY = ...
    """Multiplication f(x) = f1(x)*f2(x)."""
    DIVIDE = ...
    """Division f(x) = f1(x)/f2(x)."""
    COMPOSE = ...
    """Composition f(x) = f2(f1(x))."""

class FNParallelType(IntEnum):
    """FN parallel types.

    - `REDUNDANT`:    Every process performs the computation redundantly.
    - `SYNCHRONIZED`: The first process sends the result to the rest.
    """

    REDUNDANT = ...
    """Every process performs the computation redundantly."""
    SYNCHRONIZED = ...
    """The first process sends the result to the rest."""

class FN:
    """Mathematical function object.

    FN is used to define mathematical functions that can be evaluated
    on scalars or matrices.

    See Also
    --------
    slepc.FN
    """

    Type = FNType
    CombineType = FNCombineType
    ParallelType = FNParallelType

    # --- Unary operations ---

    def __pos__(self) -> FN: ...
    def __neg__(self) -> FN: ...

    # --- In-place binary operations ---

    def __iadd__(self, other: FN | Scalar) -> FN: ...
    def __isub__(self, other: FN | Scalar) -> FN: ...
    def __imul__(self, other: FN | Scalar) -> FN: ...
    def __idiv__(self, other: FN | Scalar) -> FN: ...
    def __itruediv__(self, other: FN | Scalar) -> FN: ...

    # --- Binary operations ---

    def __add__(self, other: FN | Scalar) -> FN: ...
    def __radd__(self, other: FN | Scalar) -> FN: ...
    def __sub__(self, other: FN | Scalar) -> FN: ...
    def __rsub__(self, other: FN | Scalar) -> FN: ...
    def __mul__(self, other: FN | Scalar) -> FN: ...
    def __rmul__(self, other: FN | Scalar) -> FN: ...
    def __div__(self, other: FN | Scalar) -> FN: ...
    def __rdiv__(self, other: FN | Scalar) -> FN: ...
    def __truediv__(self, other: FN | Scalar) -> FN: ...
    def __rtruediv__(self, other: FN | Scalar) -> FN: ...
    def __matmul__(self, other: FN) -> FN: ...

    # --- Callable ---

    def __call__(self, arg: Scalar | Mat) -> Scalar | Mat: ...

    # --- View and Lifecycle ---

    def view(self, viewer: Viewer | None = None) -> None:
        """Print the FN data structure.

        Collective.

        Parameters
        ----------
        viewer
            Visualization context; if not provided, the standard
            output is used.

        See Also
        --------
        slepc.FNView
        """
        ...

    def destroy(self) -> Self:
        """Destroy the FN object.

        Collective.

        See Also
        --------
        slepc.FNDestroy
        """
        ...

    def create(self, comm: Comm | None = None) -> Self:
        """Create the FN object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator; if not provided, it defaults to all processes.

        See Also
        --------
        slepc.FNCreate
        """
        ...

    # --- Type and Options ---

    def setType(self, fn_type: FNType | str) -> None:
        """Set the type for the FN object.

        Logically collective.

        Parameters
        ----------
        fn_type
            The function type to be used.

        See Also
        --------
        slepc.FNSetType
        """
        ...

    def getType(self) -> str:
        """Get the FN type of this object.

        Not collective.

        Returns
        -------
        str
            The function type currently being used.

        See Also
        --------
        slepc.FNGetType
        """
        ...

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """Set the prefix used for searching for all FN options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all FN option requests.

        Notes
        -----
        A hyphen (``-``) must NOT be given at the beginning of the
        prefix name.  The first character of all runtime options is
        AUTOMATICALLY the hyphen.

        See Also
        --------
        slepc.FNSetOptionsPrefix
        """
        ...

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """Append to the prefix used for searching for all FN options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all FN option requests.

        See Also
        --------
        slepc.FNAppendOptionsPrefix
        """
        ...

    def getOptionsPrefix(self) -> str:
        """Get the prefix used for searching for all FN options in the database.

        Not collective.

        Returns
        -------
        str
            The prefix string set for this FN object.

        See Also
        --------
        slepc.FNGetOptionsPrefix
        """
        ...

    def setFromOptions(self) -> None:
        """Set FN options from the options database.

        Collective.

        Notes
        -----
        To see all options, run your program with the ``-help``
        option.

        See Also
        --------
        slepc.FNSetFromOptions
        """
        ...

    def duplicate(self, comm: Comm | None = None) -> FN:
        """Duplicate the FN object copying all parameters.

        Collective.

        Duplicate the FN object copying all parameters, possibly with a
        different communicator.

        Parameters
        ----------
        comm
            MPI communicator; if not provided, it defaults to the
            object's communicator.

        See Also
        --------
        slepc.FNDuplicate
        """
        ...

    # --- Evaluation ---

    def evaluateFunction(self, x: Scalar) -> Scalar:
        """Compute the value of the function f(x) for a given x.

        Not collective.

        Parameters
        ----------
        x
            Value where the function must be evaluated.

        Returns
        -------
        Scalar
            The result of f(x).

        See Also
        --------
        slepc.FNEvaluateFunction
        """
        ...

    def evaluateDerivative(self, x: Scalar) -> Scalar:
        """Compute the value of the derivative f'(x) for a given x.

        Not collective.

        Parameters
        ----------
        x
            Value where the derivative must be evaluated.

        Returns
        -------
        Scalar
            The result of f'(x).

        See Also
        --------
        slepc.FNEvaluateDerivative
        """
        ...

    def evaluateFunctionMat(self, A: Mat, B: Mat | None = None) -> Mat:
        """Compute the value of the function f(A) for a given matrix A.

        Logically collective.

        Parameters
        ----------
        A
            Matrix on which the function must be evaluated.
        B
            Placeholder for the result.

        Returns
        -------
        Mat
            The result of f(A).

        See Also
        --------
        slepc.FNEvaluateFunctionMat
        """
        ...

    def evaluateFunctionMatVec(self, A: Mat, v: Vec | None = None) -> Vec:
        """Compute the first column of the matrix f(A) for a given matrix A.

        Logically collective.

        Parameters
        ----------
        A
            Matrix on which the function must be evaluated.
        v
            Placeholder for the result vector.

        Returns
        -------
        Vec
            The first column of the result f(A).

        See Also
        --------
        slepc.FNEvaluateFunctionMatVec
        """
        ...

    # --- Scale ---

    def setScale(self, alpha: Scalar | None = None, beta: Scalar | None = None) -> None:
        """Set the scaling parameters that define the mathematical function.

        Logically collective.

        Parameters
        ----------
        alpha
            Inner scaling (argument), default is 1.0.
        beta
            Outer scaling (result), default is 1.0.

        See Also
        --------
        slepc.FNSetScale
        """
        ...

    def getScale(self) -> tuple[Scalar, Scalar]:
        """Get the scaling parameters that define the mathematical function.

        Not collective.

        Returns
        -------
        alpha : Scalar
            Inner scaling (argument).
        beta : Scalar
            Outer scaling (result).

        See Also
        --------
        slepc.FNGetScale
        """
        ...

    # --- Method ---

    def setMethod(self, meth: int) -> None:
        """Set the method to be used to evaluate functions of matrices.

        Logically collective.

        Parameters
        ----------
        meth
            An index identifying the method.

        Notes
        -----
        In some `FN` types there are more than one algorithms available
        for computing matrix functions. In that case, this function allows
        choosing the wanted method.

        If ``meth`` is currently set to 0 and the input argument of
        `FN.evaluateFunctionMat()` is a symmetric/Hermitian matrix, then
        the computation is done via the eigendecomposition, rather than
        with the general algorithm.

        See Also
        --------
        slepc.FNSetMethod
        """
        ...

    def getMethod(self) -> int:
        """Get the method currently used for matrix functions.

        Not collective.

        Returns
        -------
        int
            An index identifying the method.

        See Also
        --------
        slepc.FNGetMethod
        """
        ...

    # --- Parallel ---

    def setParallel(self, pmode: FNParallelType) -> None:
        """Set the mode of operation in parallel runs.

        Logically collective.

        Parameters
        ----------
        pmode
            The parallel mode.

        See Also
        --------
        slepc.FNSetParallel
        """
        ...

    def getParallel(self) -> FNParallelType:
        """Get the mode of operation in parallel runs.

        Not collective.

        Returns
        -------
        FNParallelType
            The parallel mode.

        See Also
        --------
        slepc.FNGetParallel
        """
        ...

    # --- Rational Function Specific ---

    def setRationalNumerator(self, alpha: Sequence[Scalar]) -> None:
        """Set the coefficients of the numerator of the rational function.

        Logically collective.

        Parameters
        ----------
        alpha
            Coefficients.

        See Also
        --------
        slepc.FNRationalSetNumerator
        """
        ...

    def getRationalNumerator(self) -> ArrayScalar:
        """Get the coefficients of the numerator of the rational function.

        Not collective.

        Returns
        -------
        ArrayScalar
            Coefficients.

        See Also
        --------
        slepc.FNRationalGetNumerator
        """
        ...

    def setRationalDenominator(self, alpha: Sequence[Scalar]) -> None:
        """Set the coefficients of the denominator of the rational function.

        Logically collective.

        Parameters
        ----------
        alpha
            Coefficients.

        See Also
        --------
        slepc.FNRationalSetDenominator
        """
        ...

    def getRationalDenominator(self) -> ArrayScalar:
        """Get the coefficients of the denominator of the rational function.

        Not collective.

        Returns
        -------
        ArrayScalar
            Coefficients.

        See Also
        --------
        slepc.FNRationalGetDenominator
        """
        ...

    # --- Combine Function Specific ---

    def setCombineChildren(self, comb: FNCombineType, f1: FN, f2: FN) -> None:
        """Set the two child functions that constitute this combined function.

        Logically collective.

        Set the two child functions that constitute this combined function,
        and the way they must be combined.

        Parameters
        ----------
        comb
            How to combine the functions (addition, multiplication, division,
            composition).
        f1
            First function.
        f2
            Second function.

        See Also
        --------
        slepc.FNCombineSetChildren
        """
        ...

    def getCombineChildren(self) -> tuple[FNCombineType, FN, FN]:
        """Get the two child functions that constitute this combined function.

        Not collective.

        Get the two child functions that constitute this combined
        function, and the way they must be combined.

        Returns
        -------
        comb : FNCombineType
            How to combine the functions (addition, multiplication, division,
            composition).
        f1 : FN
            First function.
        f2 : FN
            Second function.

        See Also
        --------
        slepc.FNCombineGetChildren
        """
        ...

    # --- Phi Function Specific ---

    def setPhiIndex(self, k: int) -> None:
        """Set the index of the phi-function.

        Logically collective.

        Parameters
        ----------
        k
            The index.

        See Also
        --------
        slepc.FNPhiSetIndex
        """
        ...

    def getPhiIndex(self) -> int:
        """Get the index of the phi-function.

        Not collective.

        Returns
        -------
        int
            The index.

        See Also
        --------
        slepc.FNPhiGetIndex
        """
        ...

    # --- Properties ---

    @property
    def method(self) -> int:
        """The method to be used to evaluate functions of matrices."""
        ...

    @method.setter
    def method(self, value: int) -> None: ...
    @property
    def parallel(self) -> FNParallelType:
        """The mode of operation in parallel runs."""
        ...

    @parallel.setter
    def parallel(self, value: FNParallelType) -> None: ...



# -----------------------------------------------------------------------------

class BVType(object):
    """BV type."""
    MAT        = S_(BVMAT)
    SVEC       = S_(BVSVEC)
    VECS       = S_(BVVECS)
    CONTIGUOUS = S_(BVCONTIGUOUS)
    TENSOR     = S_(BVTENSOR)

class BVOrthogType(object):
    """
    BV orthogonalization types.

    - `CGS`: Classical Gram-Schmidt.
    - `MGS`: Modified Gram-Schmidt.
    """
    CGS = BV_ORTHOG_CGS
    MGS = BV_ORTHOG_MGS

class BVOrthogRefineType(object):
    """
    BV orthogonalization refinement types.

    - `IFNEEDED`: Reorthogonalize if a criterion is satisfied.
    - `NEVER`:    Never reorthogonalize.
    - `ALWAYS`:   Always reorthogonalize.
    """
    IFNEEDED = BV_ORTHOG_REFINE_IFNEEDED
    NEVER    = BV_ORTHOG_REFINE_NEVER
    ALWAYS   = BV_ORTHOG_REFINE_ALWAYS

class BVOrthogBlockType(object):
    """
    BV block-orthogonalization types.

    - `GS`:       Gram-Schmidt.
    - `CHOL`:     Cholesky.
    - `TSQR`:     Tall-skinny QR.
    - `TSQRCHOL`: Tall-skinny QR with Cholesky.
    - `SVQB`:     SVQB.
    """
    GS       = BV_ORTHOG_BLOCK_GS
    CHOL     = BV_ORTHOG_BLOCK_CHOL
    TSQR     = BV_ORTHOG_BLOCK_TSQR
    TSQRCHOL = BV_ORTHOG_BLOCK_TSQRCHOL
    SVQB     = BV_ORTHOG_BLOCK_SVQB

class BVMatMultType(object):
    """
    BV mat-mult types.

    - `VECS`: Perform a matrix-vector multiply per each column.
    - `MAT`:  Carry out a Mat-Mat product with a dense matrix.
    """
    VECS     = BV_MATMULT_VECS
    MAT      = BV_MATMULT_MAT

class BVSVDMethod(object):
    """
    BV methods for computing the SVD.

    - `REFINE`: Based on the SVD of the cross product matrix :math:`S^H S`,
                with refinement.
    - `QR`:     Based on the SVD of the triangular factor of qr(S).
    - `QR_CAA`: Variant of QR intended for use in communication-avoiding
                Arnoldi.
    """
    REFINE   = BV_SVD_METHOD_REFINE
    QR       = BV_SVD_METHOD_QR
    QR_CAA   = BV_SVD_METHOD_QR_CAA

# -----------------------------------------------------------------------------

cdef class BV(Object):

    """BV."""

    Type             = BVType
    OrthogType       = BVOrthogType
    OrthogRefineType = BVOrthogRefineType
    RefineType       = BVOrthogRefineType
    OrthogBlockType  = BVOrthogBlockType
    BlockType        = BVOrthogBlockType
    MatMultType      = BVMatMultType

    def __cinit__(self):
        self.obj = <PetscObject*> &self.bv
        self.bv = NULL

    # unary operations

    def __pos__(self):
        return bv_pos(self)

    def __neg__(self):
        return bv_neg(self)

    # inplace binary operations

    def __iadd__(self, other):
        return bv_iadd(self, other)

    def __isub__(self, other):
        return bv_isub(self, other)

    def __imul__(self, other):
        return bv_imul(self, other)

    def __idiv__(self, other):
        return bv_idiv(self, other)

    def __itruediv__(self, other):
        return bv_idiv(self, other)

    # binary operations

    def __add__(self, other):
        return bv_add(self, other)

    def __radd__(self, other):
        return bv_radd(self, other)

    def __sub__(self, other):
        return bv_sub(self, other)

    def __rsub__(self, other):
        return bv_rsub(self, other)

    def __mul__(self, other):
        return bv_mul(self, other)

    def __rmul__(self, other):
        return bv_rmul(self, other)

    def __div__(self, other):
        return bv_div(self, other)

    def __rdiv__(self, other):
        return bv_rdiv(self, other)

    def __truediv__(self, other):
        return bv_div(self, other)

    def __rtruediv__(self, other):
        return bv_rdiv(self, other)

    #

    def view(self, Viewer viewer=None) -> None:
        """
        Print the BV data structure.

        Collective.

        Parameters
        ----------
        viewer
            Visualization context; if not provided, the standard
            output is used.
        """
        cdef PetscViewer vwr = def_Viewer(viewer)
        CHKERR( BVView(self.bv, vwr) )

    def destroy(self) -> Self:
        """
        Destroy the BV object.

        Collective.
        """
        CHKERR( BVDestroy(&self.bv) )
        self.bv = NULL
        return self

    def create(self, comm: Comm | None = None) -> Self:
        """
        Create the BV object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator; if not provided, it defaults to all
            processes.
        """
        cdef MPI_Comm ccomm = def_Comm(comm, SLEPC_COMM_DEFAULT())
        cdef SlepcBV newbv = NULL
        CHKERR( BVCreate(ccomm, &newbv) )
        CHKERR( SlepcCLEAR(self.obj) ); self.bv = newbv
        return self

    def createFromMat(self, Mat A) -> Self:
        """
        Create a basis vectors object from a dense Mat object.

        Collective.

        Parameters
        ----------
        A
            A dense tall-skinny matrix.
        """
        cdef SlepcBV newbv = NULL
        CHKERR( BVCreateFromMat(A.mat, &newbv) )
        CHKERR( SlepcCLEAR(self.obj) ); self.bv = newbv
        return self

    def createMat(self) -> petsc4py.PETSc.Mat:
        """
        Create a new Mat object of dense type and copy the contents of the BV.

        Collective.

        Returns
        -------
        petsc4py.PETSc.Mat
            The new matrix.
        """
        cdef Mat mat = Mat()
        CHKERR( BVCreateMat(self.bv, &mat.mat) )
        return mat

    def duplicate(self) -> BV:
        """
        Duplicate the BV object with the same type and dimensions.

        Collective.
        """
        cdef BV bv = type(self)()
        CHKERR( BVDuplicate(self.bv, &bv.bv) )
        return bv

    def duplicateResize(self, m: int) -> BV:
        """
        Create a BV object of the same type and dimensions as an existing one.

        Collective.

        Parameters
        ----------
        m
            The number of columns.

        Notes
        -----
        With possibly different number of columns.

        """
        cdef BV bv = type(self)()
        cdef PetscInt ival = asInt(m)
        CHKERR( BVDuplicateResize(self.bv, ival, &bv.bv) )
        return bv

    def copy(self, BV result=None) -> BV:
        """
        Copy a basis vector object into another one.

        Logically collective.

        Parameters
        ----------
        result
            The copy.
        """
        if result is None:
            result = type(self)()
        if result.bv == NULL:
            CHKERR( BVDuplicate(self.bv, &result.bv) )
        CHKERR( BVCopy(self.bv, result.bv) )
        return result

    def setType(self, bv_type: Type | str) -> None:
        """
        Set the type for the BV object.

        Logically collective.

        Parameters
        ----------
        bv_type
            The inner product type to be used.
        """
        cdef SlepcBVType cval = NULL
        bv_type = str2bytes(bv_type, &cval)
        CHKERR( BVSetType(self.bv, cval) )

    def getType(self) -> str:
        """
        Get the BV type of this object.

        Not collective.

        Returns
        -------
        str
            The inner product type currently being used.
        """
        cdef SlepcBVType bv_type = NULL
        CHKERR( BVGetType(self.bv, &bv_type) )
        return bytes2str(bv_type)

    def setSizes(self, sizes: LayoutSizeSpec, m: int) -> None:
        """
        Set the local and global sizes, and the number of columns.

        Collective.

        Parameters
        ----------
        sizes
            The global size ``N`` or a two-tuple ``(n, N)``
            with the local and global sizes.
        m
            The number of columns.

        Notes
        -----
        Either ``n`` or ``N`` (but not both) can be ``PETSc.DECIDE``
        or ``None`` to have it automatically set.
        """
        cdef PetscInt n=0, N=0
        cdef PetscInt ival = asInt(m)
        BV_Sizes(sizes, &n, &N)
        CHKERR( BVSetSizes(self.bv, n, N, ival) )

    def setSizesFromVec(self, Vec w, m: int) -> None:
        """
        Set the local and global sizes, and the number of columns.

        Collective.

        Local and global sizes are specified indirectly by passing a template
        vector.

        Parameters
        ----------
        w
            The template vector.
        m
            The number of columns.
        """
        cdef PetscInt ival = asInt(m)
        CHKERR( BVSetSizesFromVec(self.bv, w.vec, ival) )

    def getSizes(self) -> tuple[LayoutSizeSpec, int]:
        """
        Get the local and global sizes, and the number of columns.

        Not collective.

        Returns
        -------
        (n, N): tuple of int
            The local and global sizes
        m: int
            The number of columns.
        """
        cdef PetscInt n=0, N=0, m=0
        CHKERR( BVGetSizes(self.bv, &n, &N, &m) )
        return ((toInt(n), toInt(N)), toInt(m))

    def setLeadingDimension(self, ld: int) -> None:
        """
        Set the leading dimension.

        Not collective.

        Parameters
        ----------
        ld
            The leading dimension.
        """
        cdef PetscInt val = asInt(ld)
        CHKERR( BVSetLeadingDimension(self.bv, val) )

    def getLeadingDimension(self) -> int:
        """
        Get the leading dimension.

        Not collective.

        Returns
        -------
        int
            The leading dimension.
        """
        cdef PetscInt val = 0
        CHKERR( BVGetLeadingDimension(self.bv, &val) )
        return toInt(val)

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Set the prefix used for searching for all BV options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all BV option requests.

        Notes
        -----
        A hyphen (``-``) must NOT be given at the beginning of the
        prefix name.  The first character of all runtime options is
        AUTOMATICALLY the hyphen.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( BVSetOptionsPrefix(self.bv, cval) )

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Append to the prefix used for searching for all BV options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all BV option requests.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( BVAppendOptionsPrefix(self.bv, cval) )

    def getOptionsPrefix(self) -> str:
        """
        Get the prefix used for searching for all BV options in the database.

        Not collective.

        Returns
        -------
        str
            The prefix string set for this BV object.
        """
        cdef const char *prefix = NULL
        CHKERR( BVGetOptionsPrefix(self.bv, &prefix) )
        return bytes2str(prefix)

    def setFromOptions(self) -> None:
        """
        Set BV options from the options database.

        Collective.

        Notes
        -----
        To see all options, run your program with the ``-help``
        option.
        """
        CHKERR( BVSetFromOptions(self.bv) )

    #

    def getOrthogonalization(self) -> tuple[OrthogType, OrthogRefineType, float, OrthogBlockType]:
        """
        Get the orthogonalization settings from the BV object.

        Not collective.

        Returns
        -------
        type: OrthogType
            The type of orthogonalization technique.
        refine: OrthogRefineType
            The type of refinement.
        eta: float
            Parameter for selective refinement (used when the
            refinement type is `BV.OrthogRefineType.IFNEEDED`).
        block: OrthogBlockType
            The type of block orthogonalization .
        """
        cdef SlepcBVOrthogType val1 = BV_ORTHOG_CGS
        cdef SlepcBVOrthogRefineType val2 = BV_ORTHOG_REFINE_IFNEEDED
        cdef SlepcBVOrthogBlockType val3 = BV_ORTHOG_BLOCK_GS
        cdef PetscReal rval = PETSC_DEFAULT
        CHKERR( BVGetOrthogonalization(self.bv, &val1, &val2, &rval, &val3) )
        return (val1, val2, toReal(rval), val3)

    def setOrthogonalization(
        self,
        otype: OrthogType | None = None,
        refine: OrthogRefineType | None = None,
        eta: float | None = None,
        block: OrthogBlockType | None = None,
    ) -> None:
        """
        Set the method used for the (block-)orthogonalization of vectors.

        Logically collective.

        Ortogonalization of vectors (classical or modified Gram-Schmidt
        with or without refinement), and for the block-orthogonalization
        (simultaneous orthogonalization of a set of vectors).

        Parameters
        ----------
        otype
            The type of orthogonalization technique.
        refine
            The type of refinement.
        eta
            Parameter for selective refinement.
        block
            The type of block orthogonalization.

        Notes
        -----
        The default settings work well for most problems.

        The parameter ``eta`` should be a real value between ``0`` and
        ``1`` (or `DETERMINE`).  The value of ``eta`` is used only when
        the refinement type is `BV.OrthogRefineType.IFNEEDED`.

        When using several processors, `BV.OrthogType.MGS` is likely to
        result in bad scalability.

        If the method set for block orthogonalization is GS, then the
        computation is done column by column with the vector orthogonalization.
        """
        cdef SlepcBVOrthogType val1 = BV_ORTHOG_CGS
        cdef SlepcBVOrthogRefineType val2 = BV_ORTHOG_REFINE_IFNEEDED
        cdef SlepcBVOrthogBlockType val3 = BV_ORTHOG_BLOCK_GS
        cdef PetscReal rval = PETSC_CURRENT
        CHKERR( BVGetOrthogonalization(self.bv, &val1, &val2, NULL, &val3) )
        if otype  is not None: val1 = otype
        if refine is not None: val2 = refine
        if block  is not None: val3 = block
        if eta    is not None: rval = asReal(eta)
        CHKERR( BVSetOrthogonalization(self.bv, val1, val2, rval, val3) )

    def getMatMultMethod(self) -> MatMultType:
        """
        Get the method used for the `matMult()` operation.

        Not collective.

        Returns
        -------
        MatMultType
            The method for the `matMult()` operation.
        """
        cdef SlepcBVMatMultType val = BV_MATMULT_MAT
        CHKERR( BVGetMatMultMethod(self.bv, &val) )
        return val

    def setMatMultMethod(self, method: MatMultType) -> None:
        """
        Set the method used for the `matMult()` operation.

        Logically collective.

        Parameters
        ----------
        method
            The method for the `matMult()` operation.
        """
        cdef SlepcBVMatMultType val = method
        CHKERR( BVSetMatMultMethod(self.bv, val) )

    #

    def getMatrix(self) -> tuple[petsc4py.PETSc.Mat, bool] | tuple[None, bool]:
        """
        Get the matrix representation of the inner product.

        Not collective.

        Returns
        -------
        mat: petsc4py.PETSc.Mat
            The matrix of the inner product
        indef: bool
            Whether the matrix is indefinite
        """
        cdef Mat mat = Mat()
        cdef PetscBool indef = PETSC_FALSE
        CHKERR( BVGetMatrix(self.bv, &mat.mat, &indef) )
        if mat.mat:
            CHKERR( PetscINCREF(mat.obj) )
            return (mat, toBool(indef))
        else:
            return (None, False)

    def setMatrix(self, Mat mat or None, indef: bool = False) -> None:
        """
        Set the bilinear form to be used for inner products.

        Collective.

        Parameters
        ----------
        mat
            The matrix of the inner product.
        indef
            Whether the matrix is indefinite
        """
        cdef PetscMat m = <PetscMat>NULL if mat is None else mat.mat
        cdef PetscBool tval = PETSC_TRUE if indef else PETSC_FALSE
        CHKERR( BVSetMatrix(self.bv, m, tval) )

    def applyMatrix(self, Vec x, Vec y) -> None:
        """
        Multiply a vector with the matrix associated to the bilinear form.

        Neighbor-wise collective.

        Parameters
        ----------
        x
            The input vector.
        y
            The result vector.

        Notes
        -----
        If the bilinear form has no associated matrix this function
        copies the vector.
        """
        CHKERR( BVApplyMatrix(self.bv, x.vec, y.vec) )

    def setActiveColumns(self, l: int, k: int) -> None:
        """
        Set the columns that will be involved in operations.

        Logically collective.

        Parameters
        ----------
        l
            The leading number of columns.
        k
            The active number of columns.
        """
        cdef PetscInt ival1 = asInt(l)
        cdef PetscInt ival2 = asInt(k)
        CHKERR( BVSetActiveColumns(self.bv, ival1, ival2) )

    def getActiveColumns(self) -> tuple[int, int]:
        """
        Get the current active dimensions.

        Not collective.

        Returns
        -------
        l: int
            The leading number of columns.
        k: int
            The active number of columns.
        """
        cdef PetscInt l=0, k=0
        CHKERR( BVGetActiveColumns(self.bv, &l, &k) )
        return (toInt(l), toInt(k))

    def scaleColumn(self, j: int, alpha: Scalar) -> None:
        """
        Scale column j by alpha.

        Logically collective.

        Parameters
        ----------
        j
            column number to be scaled.
        alpha
            scaling factor.
        """
        cdef PetscInt ival = asInt(j)
        cdef PetscScalar sval = asScalar(alpha)
        CHKERR( BVScaleColumn(self.bv, ival, sval) )

    def scale(self, alpha: Scalar) -> None:
        """
        Multiply the entries by a scalar value.

        Logically collective.

        Parameters
        ----------
        alpha
            scaling factor.

        Notes
        -----
        All active columns (except the leading ones) are scaled.
        """
        cdef PetscScalar sval = asScalar(alpha)
        CHKERR( BVScale(self.bv, sval) )

    def insertVec(self, j: int, Vec w) -> None:
        """
        Insert a vector into the specified column.

        Logically collective.

        Parameters
        ----------
        j
            The column to be overwritten.
        w
            The vector to be copied.
        """
        cdef PetscInt ival = asInt(j)
        CHKERR( BVInsertVec(self.bv, ival, w.vec) )

    def insertVecs(self, s: int, W: Vec | list[Vec], orth: bool = False) -> int:
        """
        Insert a set of vectors into specified columns.

        Collective.

        Parameters
        ----------
        s
            The first column to be overwritten.
        W
            Set of vectors to be copied.
        orth
            Flag indicating if the vectors must be orthogonalized.

        Returns
        -------
        int
            Number of linearly independent vectors.

        Notes
        -----
        Copies the contents of vectors W into self(:,s:s+n), where n is the
        length of W. If orthogonalization flag is set then the vectors are
        copied one by one then orthogonalized against the previous one.  If any
        are linearly dependent then it is discared and the value of m is
        decreased.
        """
        if isinstance(W, Vec): W = [W]
        cdef PetscInt ival = asInt(s)
        cdef PetscVec *ws = NULL
        cdef Py_ssize_t i = 0, ns = len(W)
        cdef tmp = allocate(<size_t>ns*sizeof(PetscVec),<void**>&ws)
        for i in range(ns): ws[i] = (<Vec?>W[i]).vec
        cdef PetscInt m = <PetscInt>ns
        cdef PetscBool tval = PETSC_TRUE if orth else PETSC_FALSE
        CHKERR( BVInsertVecs(self.bv, ival, &m, ws, tval) )
        return toInt(m)

    def insertConstraints(self, C: Vec | list[Vec]) -> int:
        """
        Insert a set of vectors as constraints.

        Collective.

        Parameters
        ----------
        C
            Set of vectors to be inserted as constraints.

        Returns
        -------
        int
            Number of constraints.

        Notes
        -----
        The constraints are relevant only during orthogonalization. Constraint
        vectors span a subspace that is deflated in every orthogonalization
        operation, so they are intended for removing those directions from the
        orthogonal basis computed in regular BV columns.
        """
        if isinstance(C, Vec): C = [C]
        cdef PetscVec *cs = NULL
        cdef Py_ssize_t i = 0, nc = len(C)
        cdef tmp = allocate(<size_t>nc*sizeof(PetscVec),<void**>&cs)
        for i in range(nc): cs[i] = (<Vec?>C[i]).vec
        cdef PetscInt m = <PetscInt>nc
        CHKERR( BVInsertConstraints(self.bv, &m, cs) )
        return toInt(m)

    def setNumConstraints(self, nc: int) -> None:
        """
        Set the number of constraints.

        Logically collective.

        Parameters
        ----------
        nc
            The number of constraints.
        """
        cdef PetscInt val = asInt(nc)
        CHKERR( BVSetNumConstraints(self.bv, val) )

    def getNumConstraints(self) -> int:
        """
        Get the number of constraints.

        Not collective.

        Returns
        -------
        int
            The number of constraints.
        """
        cdef PetscInt val = 0
        CHKERR( BVGetNumConstraints(self.bv, &val) )
        return toInt(val)

    def createVec(self) -> petsc4py.PETSc.Vec:
        """
        Create a Vec with the type and dimensions of the columns of the BV.

        Collective.

        Returns
        -------
        petsc4py.PETSc.Vec
            New vector.
        """
        cdef Vec v = Vec()
        CHKERR( BVCreateVec(self.bv, &v.vec) )
        return v

    def setVecType(self, vec_type: petsc4py.PETSc.Vec.Type | str) -> None:
        """
        Set the vector type.

        Collective.

        Parameters
        ----------
        vec_type
            Vector type used when creating vectors with `createVec`.
        """
        cdef PetscVecType cval = NULL
        vec_type = str2bytes(vec_type, &cval)
        CHKERR( BVSetVecType(self.bv, cval) )

    def getVecType(self) -> str:
        """
        Get the vector type used by the basis vectors object.

        Not collective.
        """
        cdef PetscVecType cval = NULL
        CHKERR( BVGetVecType(self.bv, &cval) )
        return bytes2str(cval)

    def copyVec(self, j: int, Vec v) -> None:
        """
        Copy one of the columns of a basis vectors object into a Vec.

        Logically collective.

        Parameters
        ----------
        j
            The column number to be copied.
        v
            A vector.
        """
        cdef PetscInt ival = asInt(j)
        CHKERR( BVCopyVec(self.bv, ival, v.vec) )

    def copyColumn(self, j: int, i: int) -> None:
        """
        Copy the values from one of the columns to another one.

        Logically collective.

        Parameters
        ----------
        j
            The number of the source column.
        i
            The number of the destination column.
        """
        cdef PetscInt ival1 = asInt(j)
        cdef PetscInt ival2 = asInt(i)
        CHKERR( BVCopyColumn(self.bv, ival1, ival2) )

    def setDefiniteTolerance(self, deftol: float) -> None:
        """
        Set the tolerance to be used when checking a definite inner product.

        Logically collective.

        Parameters
        ----------
        deftol
            The tolerance.
        """
        cdef PetscReal val = asReal(deftol)
        CHKERR( BVSetDefiniteTolerance(self.bv, val) )

    def getDefiniteTolerance(self) -> float:
        """
        Get the tolerance to be used when checking a definite inner product.

        Not collective.

        Returns
        -------
        float
            The tolerance.
        """
        cdef PetscReal val = 0
        CHKERR( BVGetDefiniteTolerance(self.bv, &val) )
        return toReal(val)

    def dotVec(self, Vec v) -> ArrayScalar:
        """
        Dot products of a vector against all the column vectors of the BV.

        Collective.

        Parameters
        ----------
        v
            A vector.

        Returns
        -------
        ArrayScalar 
            The computed values.

        Notes
        -----
        This is analogue to VecMDot(), but using BV to represent a collection
        of vectors. The result is :math:`m = X^H y`, so :math:`m_i` is
        equal to :math:`x_j^H y`. Note that here :math:`X` is transposed
        as opposed to BVDot().

        If a non-standard inner product has been specified with BVSetMatrix(),
        then the result is :math:`m = X^H B y`.
        """
        l, k = self.getActiveColumns()
        cdef PetscScalar* mval = NULL
        cdef tmp = allocate(<size_t>(k - l)*sizeof(PetscScalar), <void**>&mval)

        CHKERR( BVDotVec(self.bv, v.vec, mval) )

        cdef object m = None
        m = array_s(k - l, mval)
        return m

    def dotColumn(self, j: int) -> ArrayScalar:
        """
        Dot products of a column against all the column vectors of a BV.

        Collective.

        Parameters
        ----------
        j
            The index of the column.

        Returns
        -------
        ArrayScalar
            The computed values.
        """
        cdef PetscInt ival = asInt(j)
        l, k = self.getActiveColumns()
        cdef PetscScalar* mval = NULL
        cdef tmp = allocate(<size_t>(k - l)*sizeof(PetscScalar), <void**>&mval)

        CHKERR( BVDotColumn(self.bv, ival, mval) )

        cdef object m = None
        m = array_s(k - l, mval)
        return m

    def getColumn(self, j: int) -> petsc4py.PETSc.Vec:
        """
        Get a Vec object with the entries of the column of the BV object.

        Logically collective.

        Parameters
        ----------
        j
            The index of the requested column.

        Returns
        -------
        petsc4py.PETSc.Vec
            The vector containing the jth column.

        Notes
        -----
        Modifying the returned Vec will change the BV entries as well.
        """
        cdef Vec v = Vec()
        cdef PetscInt ival = asInt(j)
        CHKERR( BVGetColumn(self.bv, j, &v.vec) )
        CHKERR( PetscINCREF(v.obj) )
        return v

    def restoreColumn(self, j: int, Vec v) -> None:
        """
        Restore a column obtained with `getColumn()`.

        Logically collective.

        Parameters
        ----------
        j
            The index of the requested column.
        v
            The vector obtained with `getColumn()`.

        Notes
        -----
        The arguments must match the corresponding call to `getColumn()`.
        """
        cdef PetscInt ival = asInt(j)
        CHKERR( PetscObjectDereference(<PetscObject>v.vec) )
        CHKERR( BVRestoreColumn(self.bv, ival, &v.vec) )

    def getMat(self) -> petsc4py.PETSc.Mat:
        """
        Get a Mat object of dense type that shares the memory of the BV object.

        Collective.

        Returns
        -------
        petsc4py.PETSc.Mat
            The matrix.

        Notes
        -----
        The returned matrix contains only the active columns. If the content
        of the Mat is modified, these changes are also done in the BV object.
        The user must call `restoreMat()` when no longer needed.
        """
        cdef Mat A = Mat()
        CHKERR( BVGetMat(self.bv, &A.mat) )
        CHKERR( PetscINCREF(A.obj) )
        return A

    def restoreMat(self, Mat A) -> None:
        """
        Restore the Mat obtained with `getMat()`.

        Logically collective.

        Parameters
        ----------
        A
            The matrix obtained with `getMat()`.

        Notes
        -----
        A call to this function must match a previous call of `getMat()`.
        The effect is that the contents of the Mat are copied back to the
        BV internal data structures.
        """
        CHKERR( PetscObjectDereference(<PetscObject>A.mat) )
        CHKERR( BVRestoreMat(self.bv, &A.mat) )

    def dot(self, BV Y) -> petsc4py.PETSc.Mat:
        """
        Compute the 'block-dot' product of two basis vectors objects.

        Collective.

        :math:`M = Y^H X` :math:`(m_{ij} = y_i^H x_j)` or
        :math:`M = Y^H B X`

        Parameters
        ----------
        Y
            Left basis vectors, can be the same as self, giving
            :math:`M = X^H X`.

        Returns
        -------
        petsc4py.PETSc.Mat
            The resulting matrix.

        Notes
        -----
        This is the generalization of VecDot() for a collection of vectors,
        :math:`M = Y^H X`. The result is a matrix :math:`M` whose entry
        :math:`m_{ij}` is equal to :math:`y_i^H x_j`
        (where :math:`y_i^H` denotes the conjugate transpose of :math:`y_i`).

        :math:`X` and :math:`Y` can be the same object.

        If a non-standard inner product has been specified with setMatrix(),
        then the result is :math:`M = Y^H B X`. In this case, both
        :math:`X` and :math:`Y` must have the same associated matrix.

        Only rows (resp. columns) of :math:`M` starting from :math:`ly` (resp.
        :math:`lx`) are computed, where :math:`ly` (resp. :math:`lx`) is the
        number of leading columns of :math:`Y` (resp. :math:`X`).
        """
        cdef BV X = self
        cdef PetscInt ky=0, kx=0
        CHKERR( BVGetActiveColumns(Y.bv, NULL, &ky) )
        CHKERR( BVGetActiveColumns(X.bv, NULL, &kx) )
        cdef Mat M = Mat().createDense((ky, kx), comm=COMM_SELF).setUp()
        CHKERR( BVDot(X.bv, Y.bv, M.mat) )
        return M

    def matProject(self, Mat A: petsc4py.PETSc.Mat | None, BV Y) -> petsc4py.PETSc.Mat:
        """
        Compute the projection of a matrix onto a subspace.

        Collective.

        :math:`M = Y^H A X`

        Parameters
        ----------
        A
            Matrix to be projected.
        Y
            Left basis vectors, can be the same as self, giving
            :math:`M = X^H A X`.

        Returns
        -------
        petsc4py.PETSc.Mat
            Projection of the matrix A onto the subspace.
        """
        cdef BV X = self
        cdef PetscInt  kx=0, ky=0
        CHKERR( BVGetActiveColumns(X.bv, NULL, &kx) )
        CHKERR( BVGetActiveColumns(Y.bv, NULL, &ky) )
        cdef PetscMat Amat = <PetscMat>NULL if A is None else A.mat
        cdef Mat M = Mat().createDense((ky, kx), comm=COMM_SELF).setUp()
        CHKERR( BVMatProject(X.bv, Amat, Y.bv, M.mat) )
        return M

    def matMult(self, Mat A, BV Y=None) -> BV:
        """
        Compute the matrix-vector product for each column, :math:`Y = A V`.

        Neighbor-wise collective.

        Parameters
        ----------
        A
            The matrix.

        Returns
        -------
        BV
            The result.

        Notes
        -----
        Only active columns (excluding the leading ones) are processed.

        It is possible to choose whether the computation is done column by column
        or using dense matrices using the options database keys:

            -bv_matmult_vecs
            -bv_matmult_mat

        The default is bv_matmult_mat.
        """
        cdef MPI_Comm comm = PetscObjectComm(<PetscObject>self.bv)
        cdef SlepcBVType bv_type = NULL
        cdef PetscInt n=0, N=0, m=0
        cdef SlepcBVOrthogType val1 = BV_ORTHOG_CGS
        cdef SlepcBVOrthogRefineType val2 = BV_ORTHOG_REFINE_IFNEEDED
        cdef SlepcBVOrthogBlockType val3 = BV_ORTHOG_BLOCK_GS
        cdef PetscReal rval = PETSC_DEFAULT
        if Y is None: Y = BV()
        if Y.bv == NULL:
            CHKERR( BVGetType(self.bv, &bv_type) )
            CHKERR( MatGetLocalSize(A.mat, &n, NULL) )
            CHKERR( MatGetSize(A.mat, &N, NULL) )
            CHKERR( BVGetSizes(self.bv, NULL, NULL, &m) )
            CHKERR( BVGetOrthogonalization(self.bv, &val1, &val2, &rval, &val3) )
            CHKERR( BVCreate(comm, &Y.bv) )
            CHKERR( BVSetType(Y.bv, bv_type) )
            CHKERR( BVSetSizes(Y.bv, n, N, m) )
            CHKERR( BVSetOrthogonalization(Y.bv, val1, val2, rval, val3) )
        CHKERR( BVMatMult(self.bv, A.mat, Y.bv) )
        return Y

    def matMultHermitianTranspose(self, Mat A, BV Y=None) -> BV:
        """
        Pre-multiplication with the conjugate transpose of a matrix.

        Neighbor-wise collective.

        :math:`Y = A^H V`.

        Parameters
        ----------
        A
            The matrix.

        Returns
        -------
        BV
            The result.

        Notes
        -----
        Only active columns (excluding the leading ones) are processed.

        As opoosed to matMult(), this operation is always done by column by
        column, with a sequence of calls to MatMultHermitianTranspose().
        """
        cdef MPI_Comm comm = PetscObjectComm(<PetscObject>self.bv)
        cdef SlepcBVType bv_type = NULL
        cdef PetscInt n=0, N=0, m=0
        cdef SlepcBVOrthogType val1 = BV_ORTHOG_CGS
        cdef SlepcBVOrthogRefineType val2 = BV_ORTHOG_REFINE_IFNEEDED
        cdef SlepcBVOrthogBlockType val3 = BV_ORTHOG_BLOCK_GS
        cdef PetscReal rval = PETSC_DEFAULT
        if Y is None: Y = BV()
        if Y.bv == NULL:
            CHKERR( BVGetType(self.bv, &bv_type) )
            CHKERR( MatGetLocalSize(A.mat, NULL, &n) )
            CHKERR( MatGetSize(A.mat, NULL, &N) )
            CHKERR( BVGetSizes(self.bv, NULL, NULL, &m) )
            CHKERR( BVGetOrthogonalization(self.bv, &val1, &val2, &rval, &val3) )
            CHKERR( BVCreate(comm, &Y.bv) )
            CHKERR( BVSetType(Y.bv, bv_type) )
            CHKERR( BVSetSizes(Y.bv, n, N, m) )
            CHKERR( BVSetOrthogonalization(Y.bv, val1, val2, rval, val3) )
        CHKERR( BVMatMultHermitianTranspose(self.bv, A.mat, Y.bv) )
        return Y

    def matMultColumn(self, Mat A, j: int) -> None:
        """
        Mat-vec product for a column, storing the result in the next column.

        Neighbor-wise collective.

        :math:`v_{j+1} = A v_j`.

        Parameters
        ----------
        A
            The matrix.
        j
            Index of column.
        """
        cdef PetscInt ival = asInt(j)
        CHKERR( BVMatMultColumn(self.bv, A.mat, ival) )

    def matMultTransposeColumn(self, Mat A, j: int) -> None:
        """
        Transpose matrix-vector product for a specified column.

        Neighbor-wise collective.

        Store the result in the next column: :math:`v_{j+1} = A^T v_j`.

        Parameters
        ----------
        A
            The matrix.
        j
            Index of column.
        """
        cdef PetscInt ival = asInt(j)
        CHKERR( BVMatMultTransposeColumn(self.bv, A.mat, ival) )

    def matMultHermitianTransposeColumn(self, Mat A, j: int) -> None:
        """
        Conjugate-transpose matrix-vector product for a specified column.

        Neighbor-wise collective.

        Store the result in the next column: :math:`v_{j+1} = A^H v_j`.

        Parameters
        ----------
        A
            The matrix.
        j
            Index of column.
        """
        cdef PetscInt ival = asInt(j)
        CHKERR( BVMatMultHermitianTransposeColumn(self.bv, A.mat, ival) )

    def mult(self, alpha: Scalar, beta: Scalar, BV X, Mat Q or None) -> None:
        r"""
        Compute :math:`Y = beta Y + alpha X Q`.

        Logically collective.

        Parameters
        ----------
        alpha
            Coefficient that multiplies X.
        beta
            Coefficient that multiplies Y.
        X
            Input basis vectors.
        Q
            Input matrix, if not given the identity matrix is assumed.
        """
        cdef PetscScalar sval1 = asScalar(alpha)
        cdef PetscScalar sval2 = asScalar(beta)
        cdef PetscMat Qmat = <PetscMat>NULL if Q is None else Q.mat
        CHKERR( BVMult(self.bv, sval1, sval2, X.bv, Qmat) )

    def multInPlace(self, Mat Q, s: int, e: int) -> None:
        """
        Update a set of vectors as :math:`V(:,s:e-1) = V Q(:,s:e-1)`.

        Logically collective.

        Parameters
        ----------
        Q
            A sequential dense matrix.
        s
            First column to be overwritten.
        e
            Last column to be overwritten.
        """
        cdef PetscInt ival1 = asInt(s)
        cdef PetscInt ival2 = asInt(e)
        CHKERR( BVMultInPlace(self.bv, Q.mat, ival1, ival2) )

    def multColumn(self, alpha: Scalar, beta: Scalar, j: int, q: Sequence[Scalar]) -> None:
        r"""
        Compute :math:`y = beta y + alpha X q`.

        Logically collective.

        Compute :math:`y = beta y + alpha X q`, where
        :math:`y` is the :math:`j^{th}` column.

        Parameters
        ----------
        alpha
            Coefficient that multiplies X.
        beta
            Coefficient that multiplies y.
        j
            The column index.
        q
            Input coefficients.
        """
        cdef PetscScalar sval1 = asScalar(alpha)
        cdef PetscScalar sval2 = asScalar(beta)
        cdef PetscInt ival = asInt(j)
        cdef PetscInt nq = 0
        cdef PetscScalar* qval = NULL
        cdef tmp = iarray_s(q, &nq, &qval)
        cdef PetscInt l=0, k=0
        CHKERR( BVGetActiveColumns(self.bv, &l, &k) )
        assert nq == k-l
        CHKERR( BVMultColumn(self.bv, sval1, sval2, ival, qval) )

    def multVec(self, alpha: Scalar, beta: Scalar, Vec y, q: Sequence[Scalar]) -> None:
        """
        Compute :math:`y = beta y + alpha X q`.

        Logically collective.

        Parameters
        ----------
        alpha
            Coefficient that multiplies X.
        beta
            Coefficient that multiplies y.
        y
            Input/output vector.
        q
            Input coefficients.
        """
        cdef PetscScalar sval1 = asScalar(alpha)
        cdef PetscScalar sval2 = asScalar(beta)
        cdef PetscInt nq = 0
        cdef PetscScalar* qval = NULL
        cdef tmp = iarray_s(q, &nq, &qval)
        cdef PetscInt l=0, k=0
        CHKERR( BVGetActiveColumns(self.bv, &l, &k) )
        assert nq == k-l
        CHKERR( BVMultVec(self.bv, sval1, sval2, y.vec, qval) )

    def normColumn(self, j: int, norm_type: NormType | None = None) -> float:
        """
        Compute the vector norm of a selected column.

        Collective.

        Parameters
        ----------
        j
            Index of column.
        norm_type
            The norm type.

        Returns
        -------
        float
            The norm.

        Notes
        -----
        The norm of :math:`V_j` is computed (NORM_1, NORM_2, or NORM_INFINITY).

        If a non-standard inner product has been specified with BVSetMatrix(),
        then the returned value is :math:`\sqrt{V_j^H B V_j}`,
        where :math:`B` is the inner product matrix (argument 'type' is
        ignored).
        """
        cdef PetscNormType ntype = PETSC_NORM_2
        if norm_type is not None: ntype = norm_type
        cdef PetscReal norm = 0
        CHKERR( BVNormColumn(self.bv, j, ntype, &norm) )
        return toReal(norm)

    def norm(self, norm_type: NormType | None = None) -> float:
        """
        Compute the matrix norm of the BV.

        Collective.

        Parameters
        ----------
        norm_type
            The norm type.

        Returns
        -------
        float
            The norm.

        Notes
        -----
        All active columns (except the leading ones) are considered as a
        matrix. The allowed norms are NORM_1, NORM_FROBENIUS, and
        NORM_INFINITY.

        This operation fails if a non-standard inner product has been specified
        with BVSetMatrix().
        """
        cdef PetscNormType ntype = PETSC_NORM_FROBENIUS
        if norm_type is not None: ntype = norm_type
        cdef PetscReal norm = 0
        CHKERR( BVNorm(self.bv, ntype, &norm) )
        return toReal(norm)

    def resize(self, m: int, copy: bool = True) -> None:
        """
        Change the number of columns.

        Collective.

        Parameters
        ----------
        m
            The new number of columns.
        copy
            A flag indicating whether current values should be kept.

        Notes
        -----
        Internal storage is reallocated. If copy is True, then the contents are
        copied to the leading part of the new space.
        """
        cdef PetscInt ival = asInt(m)
        cdef PetscBool tval = PETSC_TRUE if copy else PETSC_FALSE
        CHKERR( BVResize(self.bv, ival, tval) )

    def setRandom(self) -> None:
        """
        Set the active columns of the BV to random numbers.

        Logically collective.

        Notes
        -----
        All active columns (except the leading ones) are modified.
        """
        CHKERR( BVSetRandom(self.bv) )

    def setRandomNormal(self) -> None:
        """
        Set the active columns of the BV to normal random numbers.

        Logically collective.

        Notes
        -----
        All active columns (except the leading ones) are modified.
        """
        CHKERR( BVSetRandomNormal(self.bv) )

    def setRandomSign(self) -> None:
        """
        Set the entries of a BV to values 1 or -1 with equal probability.

        Logically collective.

        Notes
        -----
        All active columns (except the leading ones) are modified.
        """
        CHKERR( BVSetRandomSign(self.bv) )

    def setRandomColumn(self, j: int) -> None:
        """
        Set one column of the BV to random numbers.

        Logically collective.

        Parameters
        ----------
        j
            Column number to be set.
        """
        cdef PetscInt ival = asInt(j)
        CHKERR( BVSetRandomColumn(self.bv, ival) )

    def setRandomCond(self, condn: float) -> None:
        """
        Set the columns of a BV to random numbers.

        Logically collective.

        The generated matrix has a prescribed condition number.

        Parameters
        ----------
        condn
            Condition number.
        """
        cdef PetscReal rval = asReal(condn)
        CHKERR( BVSetRandomCond(self.bv, rval) )

    def setRandomContext(self, Random rnd) -> None:
        """
        Set the `petsc4py.PETSc.Random` object associated with the BV.

        Collective.

        To be used in operations that need random numbers.

        Parameters
        ----------
        rnd
            The random number generator context.
        """
        CHKERR( BVSetRandomContext(self.bv, rnd.rnd) )

    def getRandomContext(self) -> Random:
        """
        Get the `petsc4py.PETSc.Random` object associated with the BV.

        Collective.

        Returns
        -------
        petsc4py.PETSc.Random
            The random number generator context.
        """
        cdef Random rnd = Random()
        CHKERR( BVGetRandomContext(self.bv, &rnd.rnd) )
        CHKERR( PetscINCREF(rnd.obj) )
        return rnd

    def orthogonalizeVec(self, Vec v) -> tuple[float, bool]:
        """
        Orthogonalize a vector with respect to a set of vectors.

        Collective.

        Parameters
        ----------
        v
            Vector to be orthogonalized, modified on return.

        Returns
        -------
        norm: float
            The norm of the resulting vector.
        lindep: bool
            Flag indicating that refinement did not improve the
            quality of orthogonalization.

        Notes
        -----
        This function applies an orthogonal projector to project vector
        :math:`v` onto the orthogonal complement of the span of the columns
        of the BV.

        This routine does not normalize the resulting vector.
        """
        cdef PetscReal norm = 0
        cdef PetscBool ldep = PETSC_FALSE
        CHKERR( BVOrthogonalizeVec(self.bv, v.vec, NULL, &norm, &ldep) )
        return (toReal(norm), toBool(ldep))

    def orthogonalizeColumn(self, j: int) -> tuple[float, bool]:
        """
        Orthogonalize a column vector with respect to the previous ones.

        Collective.

        Parameters
        ----------
        j
            Index of the column to be orthogonalized.

        Returns
        -------
        norm: float
            The norm of the resulting vector.
        lindep: bool
            Flag indicating that refinement did not improve the
            quality of orthogonalization.

        Notes
        -----
        This function applies an orthogonal projector to project vector
        :math:`V_j` onto the orthogonal complement of the span of the columns
        :math:`V[0..j-1]`, where :math:`V[.]` are the vectors of the BV.
        The columns :math:`V[0..j-1]` are assumed to be mutually orthonormal.

        This routine does not normalize the resulting vector.
        """
        cdef PetscInt ival = asInt(j)
        cdef PetscReal norm = 0
        cdef PetscBool ldep = PETSC_FALSE
        CHKERR( BVOrthogonalizeColumn(self.bv, ival, NULL, &norm, &ldep) )
        return (toReal(norm), toBool(ldep))

    def orthonormalizeColumn(self, j: int, replace: bool = False) -> tuple[float, bool]:
        """
        Orthonormalize a column vector with respect to the previous ones.

        Collective.

        This is equivalent to a call to `orthogonalizeColumn()` followed by a
        call to `scaleColumn()` with the reciprocal of the norm.

        Parameters
        ----------
        j
            Index of the column to be orthonormalized.
        replace
            Whether it is allowed to set the vector randomly.

        Returns
        -------
        norm: float
            The norm of the resulting vector.
        lindep: bool
            Flag indicating that refinement did not improve the
            quality of orthogonalization.
        """
        cdef PetscInt ival = asInt(j)
        cdef PetscBool bval = PETSC_FALSE
        if replace is not None: bval = asBool(replace)
        cdef PetscReal norm = 0
        cdef PetscBool ldep = PETSC_FALSE
        CHKERR( BVOrthonormalizeColumn(self.bv, ival, bval, &norm, &ldep) )
        return (toReal(norm), toBool(ldep))

    def orthogonalize(self, Mat R=None, **kargs: Any) -> None:
        """
        Orthogonalize all columns (except leading ones) (QR decomposition).

        Collective.

        Parameters
        ----------
        R
            A sequential dense matrix.

        Notes
        -----
        The output satisfies :math:`V_0 = V R` (where :math:`V_0` represent the
        input :math:`V`) and :math:`V' V = I`.
        """
        if kargs: self.setOrthogonalization(**kargs)
        cdef PetscMat Rmat = <PetscMat>NULL if R is None else R.mat
        CHKERR( BVOrthogonalize(self.bv, Rmat) )

    #

    property sizes:
        """Basis vectors local and global sizes, and the number of columns."""
        def __get__(self) -> tuple[LayoutSizeSpec, int]:
            return self.getSizes()

    property size:
        """Basis vectors global size."""
        def __get__(self) -> tuple[int, int]:
            return self.getSizes()[0][0]

    property local_size:
        """Basis vectors local size."""
        def __get__(self) -> int:
            return self.getSizes()[0][1]

    property column_size:
        """Basis vectors column size."""
        def __get__(self) -> int:
            return self.getSizes()[1]

# -----------------------------------------------------------------------------

del BVType
del BVOrthogType
del BVOrthogRefineType
del BVOrthogBlockType
del BVMatMultType

# -----------------------------------------------------------------------------



"""Typing support."""

from __future__ import annotations  # novermin
from typing import (  # novermin
    Callable,
    Sequence,
    Literal,
)
from numpy.typing import (
    NDArray,
)
from petsc4py.PETSc import (
    Vec,
    Mat,
    KSP,
)
from .SLEPc import (
    BV,
    DS,
    FN,
    RG,
    ST,
    EPS,
    PEP,
    NEP,
    SVD,
    MFN,
    LME,
)

__all__ = [
    'Scalar',
    'ArrayInt',
    'ArrayReal',
    'ArrayComplex',
    'ArrayScalar',
    'LayoutSizeSpec',
    'EPSStoppingFunction',
    'EPSArbitraryFunction',
    'EPSEigenvalueComparison',
    'EPSMonitorFunction',
    'PEPStoppingFunction',
    'PEPMonitorFunction',
    'NEPStoppingFunction',
    'NEPMonitorFunction',
    'SVDStoppingFunction',
    'SVDMonitorFunction',
    'MFNMonitorFunction',
    'LMEMonitorFunction',
]

# --- PETSc Sys ---

Scalar = float | complex
"""Scalar type.

Scalars can be either `float` or `complex` (but not both) depending on how
PETSc was configured (``./configure --with-scalar-type=real|complex``).

"""

ArrayInt = NDArray[int]
"""Array of `int`."""

ArrayReal = NDArray[float]
"""Array of `float`."""

ArrayComplex = NDArray[complex]
"""Array of `complex`."""

ArrayScalar = NDArray[Scalar]
"""Array of `Scalar` numbers."""

LayoutSizeSpec = int | tuple[int, int]
"""`int` or 2-`tuple` of `int` describing the layout sizes.

   A single `int` indicates global size.
   A `tuple` of `int` indicates ``(local_size, global_size)``.
"""

# --- EPS ---

EPSStoppingFunction = Callable[[EPS, int, int, int, int], EPS.ConvergedReason]
"""`EPS` stopping test callback."""

EPSArbitraryFunction = Callable[[Scalar, Scalar, Vec, Vec, Scalar, Scalar], [Scalar, Scalar]]
"""`EPS` arbitrary selection callback."""

EPSEigenvalueComparison = Callable[[Scalar, Scalar, Scalar, Scalar], int]
"""`EPS` eigenvalue comparison callback."""

EPSMonitorFunction = Callable[[EPS, int, int, ArrayScalar, ArrayScalar, ArrayReal, int], None]
"""`EPS` monitor callback."""

# --- PEP ---

PEPStoppingFunction = Callable[[PEP, int, int, int, int], PEP.ConvergedReason]
""":py:class:`PEP <slepc4py.SLEPc.PEP>` stopping test callback."""

PEPMonitorFunction = Callable[[PEP, int, int, ArrayScalar, ArrayScalar, ArrayReal, int], None]
""":py:class:`PEP <slepc4py.SLEPc.PEP>` monitor callback."""

# --- NEP ---

NEPStoppingFunction = Callable[[NEP, int, int, int, int], NEP.ConvergedReason]
""":py:class:`NEP <slepc4py.SLEPc.NEP>` stopping test callback."""

NEPMonitorFunction = Callable[[NEP, int, int, ArrayScalar, ArrayScalar, ArrayReal, int], None]
""":py:class:`NEP <slepc4py.SLEPc.NEP>` monitor callback."""

NEPFunction = Callable[[NEP, Scalar, Mat, Mat], None]
""":py:class:`NEP <slepc4py.SLEPc.NEP>` Function callback."""

NEPJacobian = Callable[[NEP, Scalar, Mat], None]
""":py:class:`NEP <slepc4py.SLEPc.NEP>` Jacobian callback."""

# --- SVD ---

SVDStoppingFunction = Callable[[SVD, int, int, int, int], SVD.ConvergedReason]
""":py:class:`SVD <slepc4py.SLEPc.SVD>` stopping test callback."""

SVDMonitorFunction = Callable[[SVD, int, int, ArrayReal, ArrayReal, int], None]
""":py:class:`SVD <slepc4py.SLEPc.SVD>` monitor callback."""

MFNMonitorFunction = Callable[[MFN, int, float], None]
"""`MFN` monitor callback."""

LMEMonitorFunction = Callable[[LME, int, float], None]
"""`LME` monitor callback."""



# -----------------------------------------------------------------------------

class RGType(object):
    """RG type."""
    INTERVAL   = S_(RGINTERVAL)
    POLYGON    = S_(RGPOLYGON)
    ELLIPSE    = S_(RGELLIPSE)
    RING       = S_(RGRING)

class RGQuadRule(object):
    """
    RG quadrature rule for contour integral methods.

    - `TRAPEZOIDAL`: Trapezoidal rule.
    - `CHEBYSHEV`:   Chebyshev points.
    """
    TRAPEZOIDAL = EPS_CISS_QUADRULE_TRAPEZOIDAL
    CHEBYSHEV   = EPS_CISS_QUADRULE_CHEBYSHEV

# -----------------------------------------------------------------------------

cdef class RG(Object):

    """RG."""

    Type     = RGType
    QuadRule = RGQuadRule

    def __cinit__(self):
        self.obj = <PetscObject*> &self.rg
        self.rg = NULL

    def view(self, Viewer viewer=None) -> None:
        """
        Print the RG data structure.

        Collective.

        Parameters
        ----------
        viewer
            Visualization context; if not provided, the standard
            output is used.
        """
        cdef PetscViewer vwr = def_Viewer(viewer)
        CHKERR( RGView(self.rg, vwr) )

    def destroy(self) -> Self:
        """
        Destroy the RG object.

        Collective.
        """
        CHKERR( RGDestroy(&self.rg) )
        self.rg = NULL
        return self

    def create(self, comm: Comm | None = None) -> Self:
        """
        Create the RG object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator; if not provided, it defaults to all processes.
        """
        cdef MPI_Comm ccomm = def_Comm(comm, SLEPC_COMM_DEFAULT())
        cdef SlepcRG newrg = NULL
        CHKERR( RGCreate(ccomm, &newrg) )
        CHKERR( SlepcCLEAR(self.obj) ); self.rg = newrg
        return self

    def setType(self, rg_type: Type | str) -> None:
        """
        Set the type for the RG object.

        Logically collective.

        Parameters
        ----------
        rg_type
            The inner product type to be used.
        """
        cdef SlepcRGType cval = NULL
        rg_type = str2bytes(rg_type, &cval)
        CHKERR( RGSetType(self.rg, cval) )

    def getType(self) -> str:
        """
        Get the RG type of this object.

        Not collective.

        Returns
        -------
        str
            The inner product type currently being used.
        """
        cdef SlepcRGType rg_type = NULL
        CHKERR( RGGetType(self.rg, &rg_type) )
        return bytes2str(rg_type)

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Set the prefix used for searching for all RG options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all RG option requests.

        Notes
        -----
        A hyphen (``-``) must NOT be given at the beginning of the
        prefix name.  The first character of all runtime options is
        AUTOMATICALLY the hyphen.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( RGSetOptionsPrefix(self.rg, cval) )

    def getOptionsPrefix(self) -> str:
        """
        Get the prefix used for searching for all RG options in the database.

        Not collective.

        Returns
        -------
        str
            The prefix string set for this RG object.
        """
        cdef const char *prefix = NULL
        CHKERR( RGGetOptionsPrefix(self.rg, &prefix) )
        return bytes2str(prefix)

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Append to the prefix used for searching for all RG options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all RG option requests.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( RGAppendOptionsPrefix(self.rg, cval) )

    def setFromOptions(self) -> None:
        """
        Set RG options from the options database.

        Collective.

        Notes
        -----
        To see all options, run your program with the ``-help``
        option.
        """
        CHKERR( RGSetFromOptions(self.rg) )

    #

    def isTrivial(self) -> bool:
        """
        Tell whether it is the trivial region (whole complex plane).

        Not collective.

        Returns
        -------
        bool
            True if the region is equal to the whole complex plane, e.g.,
            an interval region with all four endpoints unbounded or an
            ellipse with infinite radius.
        """
        cdef PetscBool tval = PETSC_FALSE
        CHKERR( RGIsTrivial(self.rg, &tval) )
        return toBool(tval)

    def isAxisymmetric(self, vertical: bool = False) -> bool:
        """
        Determine if the region is symmetric wrt. the real or imaginary axis.

        Not collective.

        Determine if the region is symmetric with respect to the real or
        imaginary axis.

        Parameters
        ----------
        vertical
            True if symmetry must be checked against the vertical axis.

        Returns
        -------
        bool
            True if the region is axisymmetric.
        """
        cdef PetscBool val = asBool(vertical)
        cdef PetscBool tval = PETSC_FALSE
        CHKERR( RGIsAxisymmetric(self.rg, val, &tval) )
        return toBool(tval)

    def getComplement(self) -> bool:
        """
        Get the flag indicating whether the region is complemented or not.

        Not collective.

        Returns
        -------
        bool
            Whether the region is complemented or not.
        """
        cdef PetscBool tval = PETSC_FALSE
        CHKERR( RGGetComplement(self.rg, &tval) )
        return toBool(tval)

    def setComplement(self, comp: bool = True) -> None:
        """
        Set a flag to indicate that the region is the complement of the specified one.

        Logically collective.

        Parameters
        ----------
        comp
            Activate/deactivate the complementation of the region.
        """
        cdef PetscBool tval = asBool(comp)
        CHKERR( RGSetComplement(self.rg, tval) )

    def setScale(self, sfactor: float = None) -> None:
        """
        Set the scaling factor to be used.

        Logically collective.

        Set the scaling factor to be used when checking that a
        point is inside the region and when computing the contour.

        Parameters
        ----------
        sfactor
            The scaling factor (default=1).
        """
        cdef PetscReal rval = 1.0
        if sfactor is not None: rval = asReal(sfactor)
        CHKERR( RGSetScale(self.rg, rval) )

    def getScale(self) -> float:
        """
        Get the scaling factor.

        Not collective.

        Returns
        -------
        float
            The scaling factor.
        """
        cdef PetscReal rval = 0
        CHKERR( RGGetScale(self.rg, &rval) )
        return toReal(rval)

    def checkInside(self, a: Sequence[complex]) -> ArrayInt:
        """
        Determine if a set of given points are inside the region or not.

        Not collective.

        Parameters
        ----------
        a
            The coordinates of the points.

        Returns
        -------
        ArrayInt
            Computed result for each point (1=inside, 0=on the contour, -1=outside).
        """
        cdef Py_ssize_t i = 0, n = len(a)
        cdef PetscScalar *ar = NULL, *ai = NULL
        cdef PetscInt *inside = NULL
        cdef tmp1 = allocate(<size_t>n*sizeof(PetscScalar),<void**>&ar)
        cdef tmp2
        if sizeof(PetscScalar) == sizeof(PetscReal):
            tmp2 = allocate(<size_t>n*sizeof(PetscScalar),<void**>&ai)
            for i in range(n):
                ar[i] = asComplexReal(a[i])
                ai[i] = asComplexImag(a[i])
        else:
            for i in range(n): ar[i] = asScalar(a[i])
        cdef tmp3 = allocate(<size_t>n*sizeof(PetscInt),<void**>&inside)
        CHKERR( RGCheckInside(self.rg, <PetscInt>n, ar, ai, inside) )
        return array_i(<PetscInt>n, inside)

    def computeContour(self, n: int) -> list[complex]:
        """
        Compute the coordinates of several points of the contour on the region.

        Not collective.

        Compute the coordinates of several points lying on the contour
        of the region.

        Parameters
        ----------
        n
            The number of points to compute.

        Returns
        -------
        list of complex
            Computed points.
        """
        cdef PetscInt k = asInt(n), i = 0
        cdef PetscScalar *cr = NULL, *ci = NULL
        cdef tmp1 = allocate(<size_t>k*sizeof(PetscScalar),<void**>&cr)
        cdef tmp2
        if sizeof(PetscScalar) == sizeof(PetscReal):
            tmp2 = allocate(<size_t>k*sizeof(PetscScalar),<void**>&ci)
        CHKERR( RGComputeContour(self.rg, k, cr, ci) )
        if sizeof(PetscScalar) == sizeof(PetscReal):
            return [toComplex(cr[i],ci[i]) for i from 0 <= i <k]
        else:
            return [toScalar(cr[i]) for i from 0 <= i <k]

    def computeBoundingBox(self) -> tuple[float, float, float, float]:
        """
        Endpoints of a rectangle in the complex plane containing the region.

        Not collective.

        Determine the endpoints of a rectangle in the complex plane that
        contains the region.

        Returns
        -------
        a: float
            The left endpoint of the bounding box in the real axis
        b: float
            The right endpoint of the bounding box in the real axis
        c: float
            The left endpoint of the bounding box in the imaginary axis
        d: float
            The right endpoint of the bounding box in the imaginary axis
        """
        cdef PetscReal a = 0, b = 0, c = 0, d = 0
        CHKERR( RGComputeBoundingBox(self.rg, &a, &b, &c, &d) )
        return (toReal(a), toReal(b), toReal(c), toReal(d))

    def canUseConjugates(self, realmats: bool = True) -> bool:
        """
        Half of integration points can be avoided (use their conjugates).

        Not collective.

        Used in contour integral methods to determine whether half of
        integration points can be avoided (use their conjugates).

        Parameters
        ----------
        realmats
            True if the problem matrices are real.

        Returns
        -------
        bool
            Whether it is possible to use conjugates.
        """
        cdef PetscBool bval = asBool(realmats)
        cdef PetscBool tval = PETSC_FALSE
        CHKERR( RGCanUseConjugates(self.rg, bval, &tval) )
        return toBool(tval)

    def computeQuadrature(self, quad: QuadRule, n: int) -> tuple[ArrayScalar, ArrayScalar, ArrayScalar]:
        """
        Compute the values of the parameters used in a quadrature rule.

        Not collective.

        Compute the values of the parameters used in a quadrature rule for a
        contour integral around the boundary of the region.

        Parameters
        ----------
        quad
            The type of quadrature.
        n
            The number of quadrature points to compute.

        Returns
        -------
        z: ArrayScalar
            Quadrature points.
        zn: ArrayScalar
            Normalized quadrature points.
        w: ArrayScalar
            Quadrature weights.
        """
        cdef SlepcRGQuadRule val = quad
        cdef PetscInt k = asInt(n), i = 0
        cdef PetscScalar *z = NULL, *zn = NULL, *w = NULL
        cdef tmp1 = allocate(<size_t>k*sizeof(PetscScalar),<void**>&z)
        cdef tmp2 = allocate(<size_t>k*sizeof(PetscScalar),<void**>&zn)
        cdef tmp3 = allocate(<size_t>k*sizeof(PetscScalar),<void**>&w)
        CHKERR( RGComputeQuadrature(self.rg, val, k, z, zn, w) )
        return (array_s(k, z), array_s(k, zn), array_s(k, w))

    #

    def setEllipseParameters(self, center: Scalar, radius: float, vscale: float | None = None) -> None:
        """
        Set the parameters defining the ellipse region.

        Logically collective.

        Parameters
        ----------
        center
            The center.
        radius
            The radius.
        vscale
            The vertical scale.
        """
        cdef PetscScalar sval = asScalar(center)
        cdef PetscReal val1 = asReal(radius)
        cdef PetscReal val2 = 1.0
        if vscale is not None: val2 = asReal(vscale)
        CHKERR( RGEllipseSetParameters(self.rg, sval, val1, val2) )

    def getEllipseParameters(self) -> tuple[Scalar, float, float]:
        """
        Get the parameters that define the ellipse region.

        Not collective.

        Returns
        -------
        center: Scalar
            The center.
        radius: float
            The radius.
        vscale: float
            The vertical scale.
        """
        cdef PetscScalar sval = 0
        cdef PetscReal val1 = 0
        cdef PetscReal val2 = 0
        CHKERR( RGEllipseGetParameters(self.rg, &sval, &val1, &val2) )
        return (toScalar(sval), toReal(val1), toReal(val2))

    def setIntervalEndpoints(self, a: float, b: float, c: float, d: float) -> None:
        """
        Set the parameters defining the interval region.

        Logically collective.

        Parameters
        ----------
        a
            The left endpoint in the real axis.
        b
            The right endpoint in the real axis.
        c
            The upper endpoint in the imaginary axis.
        d
            The lower endpoint in the imaginary axis.
        """
        cdef PetscReal va = asReal(a)
        cdef PetscReal vb = asReal(b)
        cdef PetscReal vc = asReal(c)
        cdef PetscReal vd = asReal(d)
        CHKERR( RGIntervalSetEndpoints(self.rg, va, vb, vc, vd) )

    def getIntervalEndpoints(self) -> tuple[float, float, float, float]:
        """
        Get the parameters that define the interval region.

        Not collective.

        Returns
        -------
        a: float
            The left endpoint in the real axis.
        b: float
            The right endpoint in the real axis.
        c: float
            The upper endpoint in the imaginary axis.
        d: float
            The lower endpoint in the imaginary axis.
        """
        cdef PetscReal va = 0
        cdef PetscReal vb = 0
        cdef PetscReal vc = 0
        cdef PetscReal vd = 0
        CHKERR( RGIntervalGetEndpoints(self.rg, &va, &vb, &vc, &vd) )
        return (toReal(va), toReal(vb), toReal(vc), toReal(vd))

    def setPolygonVertices(self, v: Sequence[float] | Sequence[Scalar]) -> None:
        """
        Set the vertices that define the polygon region.

        Logically collective.

        Parameters
        ----------
        v
            The vertices.
        """
        cdef Py_ssize_t i = 0, n = len(v)
        cdef PetscScalar *vr = NULL, *vi = NULL
        cdef tmp1 = allocate(<size_t>n*sizeof(PetscScalar),<void**>&vr)
        cdef tmp2
        if sizeof(PetscScalar) == sizeof(PetscReal):
            tmp2 = allocate(<size_t>n*sizeof(PetscScalar),<void**>&vi)
            for i in range(n):
                vr[i] = asComplexReal(v[i])
                vi[i] = asComplexImag(v[i])
        else:
            for i in range(n): vr[i] = asScalar(v[i])
        CHKERR( RGPolygonSetVertices(self.rg, <PetscInt>n, vr, vi) )

    def getPolygonVertices(self) -> ArrayComplex:
        """
        Get the parameters that define the interval region.

        Not collective.

        Returns
        -------
        ArrayComplex
            The vertices.
        """
        cdef PetscInt n = 0
        cdef PetscScalar *vr = NULL, *vi = NULL
        CHKERR( RGPolygonGetVertices(self.rg, &n, &vr, &vi) )
        if sizeof(PetscScalar) == sizeof(PetscReal):
            v = [toComplex(vr[i],vi[i]) for i from 0 <= i <n]
            CHKERR( PetscFree(vi) )
        else:
            v = [toScalar(vr[i]) for i from 0 <= i <n]
        CHKERR( PetscFree(vr) )
        return v

    def setRingParameters(
        self,
        center: Scalar,
        radius: float,
        vscale: float,
        start_ang: float,
        end_ang: float,
        width: float,
    ) -> None:
        """
        Set the parameters defining the ring region.

        Logically collective.

        Parameters
        ----------
        center
            The center.
        radius
            The radius.
        vscale
            The vertical scale.
        start_ang
            The right-hand side angle.
        end_ang
            The left-hand side angle.
        width
            The width of the ring.
        """
        cdef PetscScalar sval = asScalar(center)
        cdef PetscReal val1 = asReal(radius)
        cdef PetscReal val2 = asReal(vscale)
        cdef PetscReal val3 = asReal(start_ang)
        cdef PetscReal val4 = asReal(end_ang)
        cdef PetscReal val5 = asReal(width)
        CHKERR( RGRingSetParameters(self.rg, sval, val1, val2, val3, val4, val5) )

    def getRingParameters(self) -> tuple[Scalar, float, float, float, float, float]:
        """
        Get the parameters that define the ring region.

        Not collective.

        Returns
        -------
        center: Scalar
            The center.
        radius: float
            The radius.
        vscale: float
            The vertical scale.
        start_ang: float
            The right-hand side angle.
        end_ang: float
            The left-hand side angle.
        width: float
            The width of the ring.
        """
        cdef PetscScalar sval = 0
        cdef PetscReal val1 = 0
        cdef PetscReal val2 = 0
        cdef PetscReal val3 = 0
        cdef PetscReal val4 = 0
        cdef PetscReal val5 = 0
        CHKERR( RGRingGetParameters(self.rg, &sval, &val1, &val2, &val3, &val4, &val5) )
        return (toScalar(sval), toReal(val1), toReal(val2), toReal(val3), toReal(val4), toReal(val5))

    #

    property complement:
        """If the region is the complement of the specified one."""
        def __get__(self) -> bool:
            return self.getComplement()
        def __set__(self, value):
            self.setComplement(value)

    property scale:
        """The scaling factor to be used."""
        def __get__(self) -> float:
            return self.getScale()
        def __set__(self, value):
            self.setScale(value)

# -----------------------------------------------------------------------------

del RGType
del RGQuadRule

# -----------------------------------------------------------------------------



cdef extern from * nogil:

    ctypedef char* SlepcMFNType "const char*"
    SlepcMFNType MFNKRYLOV
    SlepcMFNType MFNEXPOKIT

    ctypedef enum SlepcMFNConvergedReason "MFNConvergedReason":
        MFN_CONVERGED_TOL
        MFN_CONVERGED_ITS
        MFN_DIVERGED_ITS
        MFN_DIVERGED_BREAKDOWN
        MFN_CONVERGED_ITERATING

    ctypedef PetscErrorCode (*SlepcMFNCtxDel)(void*)
    ctypedef PetscErrorCode (*SlepcMFNMonitorFunction)(SlepcMFN,
                                            PetscInt,
                                            PetscReal,
                                            void*) except PETSC_ERR_PYTHON

    PetscErrorCode MFNCreate(MPI_Comm,SlepcMFN*)
    PetscErrorCode MFNDestroy(SlepcMFN*)
    PetscErrorCode MFNReset(SlepcMFN)
    PetscErrorCode MFNView(SlepcMFN,PetscViewer)

    PetscErrorCode MFNSetType(SlepcMFN,SlepcMFNType)
    PetscErrorCode MFNGetType(SlepcMFN,SlepcMFNType*)
    PetscErrorCode MFNSetOperator(SlepcMFN,PetscMat)
    PetscErrorCode MFNGetOperator(SlepcMFN,PetscMat*)
    PetscErrorCode MFNSetOptionsPrefix(SlepcMFN,char*)
    PetscErrorCode MFNGetOptionsPrefix(SlepcMFN,char*[])
    PetscErrorCode MFNSetFromOptions(SlepcMFN)
    PetscErrorCode MFNAppendOptionsPrefix(SlepcMFN,char*)
    PetscErrorCode MFNSetUp(SlepcMFN)
    PetscErrorCode MFNSolve(SlepcMFN,PetscVec,PetscVec)
    PetscErrorCode MFNSolveTranspose(SlepcMFN,PetscVec,PetscVec)

    PetscErrorCode MFNSetBV(SlepcMFN,SlepcBV)
    PetscErrorCode MFNGetBV(SlepcMFN,SlepcBV*)
    PetscErrorCode MFNSetFN(SlepcMFN,SlepcFN)
    PetscErrorCode MFNGetFN(SlepcMFN,SlepcFN*)
    PetscErrorCode MFNSetTolerances(SlepcMFN,PetscReal,PetscInt)
    PetscErrorCode MFNGetTolerances(SlepcMFN,PetscReal*,PetscInt*)
    PetscErrorCode MFNSetDimensions(SlepcMFN,PetscInt)
    PetscErrorCode MFNGetDimensions(SlepcMFN,PetscInt*)

    PetscErrorCode MFNSetErrorIfNotConverged(SlepcMFN,PetscBool)
    PetscErrorCode MFNGetErrorIfNotConverged(SlepcMFN,PetscBool*)

    PetscErrorCode MFNMonitorSet(SlepcMFN,SlepcMFNMonitorFunction,void*,SlepcMFNCtxDel)
    PetscErrorCode MFNMonitorCancel(SlepcMFN)
    PetscErrorCode MFNGetIterationNumber(SlepcMFN,PetscInt*)

    PetscErrorCode MFNGetConvergedReason(SlepcMFN,SlepcMFNConvergedReason*)

# -----------------------------------------------------------------------------

cdef inline MFN ref_MFN(SlepcMFN mfn):
    cdef MFN ob = <MFN> MFN()
    ob.mfn = mfn
    CHKERR( PetscINCREF(ob.obj) )
    return ob

# -----------------------------------------------------------------------------

cdef PetscErrorCode MFN_Monitor(
    SlepcMFN    mfn,
    PetscInt    it,
    PetscReal   errest,
    void        *ctx,
    ) except PETSC_ERR_PYTHON with gil:
    cdef MFN Mfn = ref_MFN(mfn)
    cdef object monitorlist = Mfn.get_attr('__monitor__')
    if monitorlist is None: return PETSC_SUCCESS
    for (monitor, args, kargs) in monitorlist:
        monitor(Mfn, toInt(it), toReal(errest), *args, **kargs)
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------

cdef class Util:
    """Util."""

    @classmethod
    def createMatBSE(cls, Mat R: petsc4py.PETSc.Mat, Mat C: petsc4py.PETSc.Mat) -> petsc4py.PETSc.Mat:
        """
        Create a matrix that can be used to define a BSE type problem.

        Collective.

        Create a matrix that can be used to define a structured eigenvalue
        problem of type BSE (Bethe-Salpeter Equation).

        Parameters
        ----------
        R
            The matrix for the diagonal block (resonant).
        C
            The matrix for the off-diagonal block (coupling).

        Returns
        -------
        petsc4py.PETSc.Mat
            The matrix with the block form :math:`H = [ R\; C; {-C}^H\; {-R}^T ]`.
        """
        cdef Mat H = Mat()
        CHKERR( MatCreateBSE(R.mat, C.mat, &H.mat) )
        return H

    @classmethod
    def createMatHamiltonian(cls, Mat A: petsc4py.PETSc.Mat, Mat B: petsc4py.PETSc.Mat, Mat C: petsc4py.PETSc.Mat) -> petsc4py.PETSc.Mat:
        """
        Create matrix to be used for a structured Hamiltonian eigenproblem.

        Collective.

        Parameters
        ----------
        A
            The matrix for (0,0) block.
        B
            The matrix for (0,1) block, must be real symmetric or Hermitian.
        C
            The matrix for (1,0) block, must be real symmetric or Hermitian.

        Returns
        -------
        petsc4py.PETSc.Mat
            The matrix with the block form :math:`H = [ A B; C -A^* ]`.
        """
        cdef Mat H = Mat()
        CHKERR( MatCreateHamiltonian(A.mat, B.mat, C.mat, &H.mat) )
        return H

# -----------------------------------------------------------------------------
